import warnings
from urllib.parse import urljoin

from rest_framework.fields import empty
from rest_framework.relations import ManyRelatedField
from rest_framework.schemas import openapi as drf_openapi
from rest_framework.schemas.utils import is_list_view

from rest_framework_json_api import serializers, views
from rest_framework_json_api.utils import format_field_name


class SchemaGenerator(drf_openapi.SchemaGenerator):
    """
    Extend DRF's SchemaGenerator to implement JSON:API flavored generateschema command.
    """

    #: These JSON:API component definitions are referenced by the generated OAS schema.
    #: If you need to add more or change these static component definitions, extend this dict.
    jsonapi_components = {
        "schemas": {
            "jsonapi": {
                "type": "object",
                "description": "The server's implementation",
                "properties": {
                    "version": {"type": "string"},
                    "meta": {"$ref": "#/components/schemas/meta"},
                },
                "additionalProperties": False,
            },
            "resource": {
                "type": "object",
                "required": ["type", "id"],
                "additionalProperties": False,
                "properties": {
                    "type": {"$ref": "#/components/schemas/type"},
                    "id": {"$ref": "#/components/schemas/id"},
                    "attributes": {
                        "type": "object",
                        # ...
                    },
                    "relationships": {
                        "type": "object",
                        # ...
                    },
                    "links": {"$ref": "#/components/schemas/links"},
                    "meta": {"$ref": "#/components/schemas/meta"},
                },
            },
            "link": {
                "oneOf": [
                    {
                        "description": "a string containing the link's URL",
                        "type": "string",
                        "format": "uri-reference",
                    },
                    {
                        "type": "object",
                        "required": ["href"],
                        "properties": {
                            "href": {
                                "description": "a string containing the link's URL",
                                "type": "string",
                                "format": "uri-reference",
                            },
                            "meta": {"$ref": "#/components/schemas/meta"},
                        },
                    },
                ]
            },
            "links": {
                "type": "object",
                "additionalProperties": {"$ref": "#/components/schemas/link"},
            },
            "reltoone": {
                "description": "a singular 'to-one' relationship",
                "type": "object",
                "properties": {
                    "links": {"$ref": "#/components/schemas/relationshipLinks"},
                    "data": {"$ref": "#/components/schemas/relationshipToOne"},
                    "meta": {"$ref": "#/components/schemas/meta"},
                },
            },
            "relationshipToOne": {
                "description": "reference to other resource in a to-one relationship",
                "anyOf": [
                    {"$ref": "#/components/schemas/nulltype"},
                    {"$ref": "#/components/schemas/linkage"},
                ],
            },
            "reltomany": {
                "description": "a multiple 'to-many' relationship",
                "type": "object",
                "properties": {
                    "links": {"$ref": "#/components/schemas/relationshipLinks"},
                    "data": {"$ref": "#/components/schemas/relationshipToMany"},
                    "meta": {"$ref": "#/components/schemas/meta"},
                },
            },
            "relationshipLinks": {
                "description": "optional references to other resource objects",
                "type": "object",
                "additionalProperties": True,
                "properties": {
                    "self": {"$ref": "#/components/schemas/link"},
                    "related": {"$ref": "#/components/schemas/link"},
                },
            },
            "relationshipToMany": {
                "description": "An array of objects each containing the "
                "'type' and 'id' for to-many relationships",
                "type": "array",
                "items": {"$ref": "#/components/schemas/linkage"},
                "uniqueItems": True,
            },
            # A RelationshipView uses a ResourceIdentifierObjectSerializer (hence the name
            # ResourceIdentifierObject returned by get_component_name()) which serializes type and
            # id. These can be lists or individual items depending on whether the relationship is
            # toMany or toOne so offer both options since we are not iterating over all the
            # possible {related_field}'s but rather rendering one path schema which may represent
            # toMany and toOne relationships.
            "ResourceIdentifierObject": {
                "oneOf": [
                    {"$ref": "#/components/schemas/relationshipToOne"},
                    {"$ref": "#/components/schemas/relationshipToMany"},
                ]
            },
            "linkage": {
                "type": "object",
                "description": "the 'type' and 'id'",
                "required": ["type", "id"],
                "properties": {
                    "type": {"$ref": "#/components/schemas/type"},
                    "id": {"$ref": "#/components/schemas/id"},
                    "meta": {"$ref": "#/components/schemas/meta"},
                },
            },
            "pagination": {
                "type": "object",
                "properties": {
                    "first": {"$ref": "#/components/schemas/pageref"},
                    "last": {"$ref": "#/components/schemas/pageref"},
                    "prev": {"$ref": "#/components/schemas/pageref"},
                    "next": {"$ref": "#/components/schemas/pageref"},
                },
            },
            "pageref": {
                "oneOf": [
                    {"type": "string", "format": "uri-reference"},
                    {"$ref": "#/components/schemas/nulltype"},
                ]
            },
            "failure": {
                "type": "object",
                "required": ["errors"],
                "properties": {
                    "errors": {"$ref": "#/components/schemas/errors"},
                    "meta": {"$ref": "#/components/schemas/meta"},
                    "jsonapi": {"$ref": "#/components/schemas/jsonapi"},
                    "links": {"$ref": "#/components/schemas/links"},
                },
            },
            "errors": {
                "type": "array",
                "items": {"$ref": "#/components/schemas/error"},
                "uniqueItems": True,
            },
            "error": {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "id": {"type": "string"},
                    "status": {"type": "string"},
                    "links": {"$ref": "#/components/schemas/links"},
                    "code": {"type": "string"},
                    "title": {"type": "string"},
                    "detail": {"type": "string"},
                    "source": {
                        "type": "object",
                        "properties": {
                            "pointer": {
                                "type": "string",
                                "description": "A [JSON Pointer](https://tools.ietf.org/html/rfc6901) "
                                "to the associated entity in the request document "
                                "[e.g. `/data` for a primary data object, or "
                                "`/data/attributes/title` for a specific attribute.",
                            },
                            "parameter": {
                                "type": "string",
                                "description": "A string indicating which query parameter "
                                "caused the error.",
                            },
                            "meta": {"$ref": "#/components/schemas/meta"},
                        },
                    },
                },
            },
            "onlymeta": {
                "additionalProperties": False,
                "properties": {"meta": {"$ref": "#/components/schemas/meta"}},
            },
            "meta": {"type": "object", "additionalProperties": True},
            "datum": {
                "description": "singular item",
                "properties": {"data": {"$ref": "#/components/schemas/resource"}},
            },
            "nulltype": {"type": "object", "nullable": True, "default": None},
            "type": {
                "type": "string",
                "description": "The [type]"
                "(https://jsonapi.org/format/#document-resource-object-identification) "
                "member is used to describe resource objects that share common attributes "
                "and relationships.",
            },
            "id": {
                "type": "string",
                "description": "Each resource object’s type and id pair MUST "
                "[identify]"
                "(https://jsonapi.org/format/#document-resource-object-identification) "
                "a single, unique resource.",
            },
        },
        "parameters": {
            "include": {
                "name": "include",
                "in": "query",
                "description": "[list of included related resources]"
                "(https://jsonapi.org/format/#fetching-includes)",
                "required": False,
                "style": "form",
                "schema": {"type": "string"},
            },
            # TODO: deepObject not well defined/supported:
            #       https://github.com/OAI/OpenAPI-Specification/issues/1706
            "fields": {
                "name": "fields",
                "in": "query",
                "description": "[sparse fieldsets]"
                "(https://jsonapi.org/format/#fetching-sparse-fieldsets).\n"
                "Use fields[\\<typename\\>]=field1,field2,...,fieldN",
                "required": False,
                "style": "deepObject",
                "schema": {
                    "type": "object",
                },
                "explode": True,
            },
            "sort": {
                "name": "sort",
                "in": "query",
                "description": "[list of fields to sort by]"
                "(https://jsonapi.org/format/#fetching-sorting)",
                "required": False,
                "style": "form",
                "schema": {"type": "string"},
            },
        },
    }

    def get_schema(self, request=None, public=False):
        """
        Generate a JSON:API OpenAPI schema.
        Overrides upstream DRF's get_schema.
        """
        # TODO: avoid copying so much of upstream get_schema()
        schema = super().get_schema(request, public)

        components_schemas = {}

        # Iterate endpoints generating per method path operations.
        paths = {}
        _, view_endpoints = self._get_paths_and_endpoints(None if public else request)

        #: `expanded_endpoints` is like view_endpoints with one extra field tacked on:
        #: - 'action' copy of current view.action (list/fetch) as this gets reset for each request.
        expanded_endpoints = []
        for path, method, view in view_endpoints:
            if hasattr(view, "action") and view.action == "retrieve_related":
                expanded_endpoints += self._expand_related(
                    path, method, view, view_endpoints
                )
            else:
                expanded_endpoints.append(
                    (path, method, view, getattr(view, "action", None))
                )

        for path, method, view, action in expanded_endpoints:
            if not self.has_view_permissions(path, method, view):
                continue
            # kludge to preserve view.action as it is 'list' for the parent ViewSet
            # but the related viewset that was expanded may be either 'fetch' (to_one) or 'list'
            # (to_many). This patches the view.action appropriately so that
            # view.schema.get_operation() "does the right thing" for fetch vs. list.
            current_action = None
            if hasattr(view, "action"):
                current_action = view.action
                view.action = action
            operation = view.schema.get_operation(path, method)
            components = view.schema.get_components(path, method)
            for k in components.keys():
                if k not in components_schemas:
                    continue
                if components_schemas[k] == components[k]:
                    continue
                warnings.warn(
                    'Schema component "{}" has been overriden with a different value.'.format(
                        k
                    )
                )

            components_schemas.update(components)

            if hasattr(view, "action"):
                view.action = current_action
            # Normalise path for any provided mount url.
            if path.startswith("/"):
                path = path[1:]
            path = urljoin(self.url or "/", path)

            paths.setdefault(path, {})
            paths[path][method.lower()] = operation

        self.check_duplicate_operation_id(paths)

        # Compile final schema, overriding stuff from super class.
        schema["paths"] = paths
        schema["components"] = self.jsonapi_components
        schema["components"]["schemas"].update(components_schemas)

        return schema

    def _expand_related(self, path, method, view, view_endpoints):
        """
        Expand path containing .../{id}/{related_field} into list of related fields
        and **their** views, making sure toOne relationship's views are a 'fetch' and toMany
        relationship's are a 'list'.
        :param path
        :param method
        :param view
        :param view_endpoints
        :return:list[tuple(path, method, view, action)]
        """
        result = []
        serializer = view.get_serializer()
        # It's not obvious if it's allowed to have both included_ and related_ serializers,
        # so just merge both dicts.
        serializers = {}
        if hasattr(serializer, "included_serializers"):
            serializers = {**serializers, **serializer.included_serializers}
        if hasattr(serializer, "related_serializers"):
            serializers = {**serializers, **serializer.related_serializers}
        related_fields = [fs for fs in serializers.items()]

        for field, related_serializer in related_fields:
            related_view = self._find_related_view(
                view_endpoints, related_serializer, view
            )
            if related_view:
                action = self._field_is_one_or_many(field, view)
                result.append(
                    (
                        path.replace("{related_field}", field),
                        method,
                        related_view,
                        action,
                    )
                )

        return result

    def _find_related_view(self, view_endpoints, related_serializer, parent_view):
        """
        For a given related_serializer, try to find it's "parent" view instance in view_endpoints.
        :param view_endpoints: list of all view endpoints
        :param related_serializer: the related serializer for a given related field
        :param parent_view: the parent view (used to find toMany vs. toOne).
               TODO: not actually used.
        :return:view
        """
        for path, method, view in view_endpoints:
            view_serializer = view.get_serializer()
            if isinstance(view_serializer, related_serializer):
                return view

        return None

    def _field_is_one_or_many(self, field, view):
        serializer = view.get_serializer()
        if isinstance(serializer.fields[field], ManyRelatedField):
            return "list"
        else:
            return "fetch"


class AutoSchema(drf_openapi.AutoSchema):
    """
    Extend DRF's openapi.AutoSchema for JSON:API serialization.
    """

    #: ignore all the media types and only generate a JSON:API schema.
    content_types = ["application/vnd.api+json"]

    def get_operation(self, path, method):
        """
        JSON:API adds some standard fields to the API response that are not in upstream DRF:
        - some that only apply to GET/HEAD methods.
        - collections
        - special handling for POST, PATCH, DELETE
        """
        operation = {}
        operation["operationId"] = self.get_operation_id(path, method)
        operation["description"] = self.get_description(path, method)

        parameters = []
        parameters += self.get_path_parameters(path, method)
        # pagination, filters only apply to GET/HEAD of collections and items
        if method in ["GET", "HEAD"]:
            parameters += self._get_include_parameters(path, method)
            parameters += self._get_fields_parameters(path, method)
            parameters += self._get_sort_parameters(path, method)
            parameters += self.get_pagination_parameters(path, method)
            parameters += self.get_filter_parameters(path, method)
        operation["parameters"] = parameters
        operation["tags"] = self.get_tags(path, method)

        # get request and response code schemas
        if method == "GET":
            if is_list_view(path, method, self.view):
                self._add_get_collection_response(operation)
            else:
                self._add_get_item_response(operation)
        elif method == "POST":
            self._add_post_item_response(operation, path)
        elif method == "PATCH":
            self._add_patch_item_response(operation, path)
        elif method == "DELETE":
            # should only allow deleting a resource, not a collection
            # TODO: implement delete of a relationship in future release.
            self._add_delete_item_response(operation, path)
        return operation

    def get_operation_id(self, path, method):
        """
        The upstream DRF version creates non-unique operationIDs, because the same view is
        used for the main path as well as such as related and relationships.
        This concatenates the (mapped) method name and path as the spec allows most any
        """
        method_name = getattr(self.view, "action", method.lower())
        if is_list_view(path, method, self.view):
            action = "List"
        elif method_name not in self.method_mapping:
            action = method_name
        else:
            action = self.method_mapping[method.lower()]
        return action + path

    def _get_include_parameters(self, path, method):
        """
        includes parameter: https://jsonapi.org/format/#fetching-includes
        """
        return [{"$ref": "#/components/parameters/include"}]

    def _get_fields_parameters(self, path, method):
        """
        sparse fieldsets https://jsonapi.org/format/#fetching-sparse-fieldsets
        """
        # TODO: See if able to identify the specific types for fields[type]=... and return this:
        # name: fields
        # in: query
        # description: '[sparse fieldsets](https://jsonapi.org/format/#fetching-sparse-fieldsets)'
        # required: true
        # style: deepObject
        # schema:
        #   type: object
        #   properties:
        #     hello:
        #       type: string  # noqa F821
        #     world:
        #       type: string  # noqa F821
        # explode: true
        return [{"$ref": "#/components/parameters/fields"}]

    def _get_sort_parameters(self, path, method):
        """
        sort parameter: https://jsonapi.org/format/#fetching-sorting
        """
        return [{"$ref": "#/components/parameters/sort"}]

    def _add_get_collection_response(self, operation):
        """
        Add GET 200 response for a collection to operation
        """
        operation["responses"] = {
            "200": self._get_toplevel_200_response(operation, collection=True)
        }
        self._add_get_4xx_responses(operation)

    def _add_get_item_response(self, operation):
        """
        add GET 200 response for an item to operation
        """
        operation["responses"] = {
            "200": self._get_toplevel_200_response(operation, collection=False)
        }
        self._add_get_4xx_responses(operation)

    def _get_toplevel_200_response(self, operation, collection=True):
        """
        return top-level JSON:API GET 200 response

        :param collection: True for collections; False for individual items.

        Uses a $ref to the components.schemas.<Name> component definition.
        """
        if collection:
            data = {
                "type": "array",
                "items": self._get_reference(self.view.get_serializer()),
            }
        else:
            data = self._get_reference(self.view.get_serializer())

        return {
            "description": operation["operationId"],
            "content": {
                "application/vnd.api+json": {
                    "schema": {
                        "type": "object",
                        "required": ["data"],
                        "properties": {
                            "data": data,
                            "included": {
                                "type": "array",
                                "uniqueItems": True,
                                "items": {"$ref": "#/components/schemas/resource"},
                            },
                            "links": {
                                "description": "Link members related to primary data",
                                "allOf": [
                                    {"$ref": "#/components/schemas/links"},
                                    {"$ref": "#/components/schemas/pagination"},
                                ],
                            },
                            "jsonapi": {"$ref": "#/components/schemas/jsonapi"},
                        },
                    }
                }
            },
        }

    def _add_post_item_response(self, operation, path):
        """
        add response for POST of an item to operation
        """
        operation["requestBody"] = self.get_request_body(path, "POST")
        operation["responses"] = {
            "201": self._get_toplevel_200_response(operation, collection=False)
        }
        operation["responses"]["201"]["description"] = (
            "[Created](https://jsonapi.org/format/#crud-creating-responses-201). "
            "Assigned `id` and/or any other changes are in this response."
        )
        self._add_async_response(operation)
        operation["responses"]["204"] = {
            "description": "[Created](https://jsonapi.org/format/#crud-creating-responses-204) "
            "with the supplied `id`. No other changes from what was POSTed."
        }
        self._add_post_4xx_responses(operation)

    def _add_patch_item_response(self, operation, path):
        """
        Add PATCH response for an item to operation
        """
        operation["requestBody"] = self.get_request_body(path, "PATCH")
        operation["responses"] = {
            "200": self._get_toplevel_200_response(operation, collection=False)
        }
        self._add_patch_4xx_responses(operation)

    def _add_delete_item_response(self, operation, path):
        """
        add DELETE response for item or relationship(s) to operation
        """
        # Only DELETE of relationships has a requestBody
        if isinstance(self.view, views.RelationshipView):
            operation["requestBody"] = self.get_request_body(path, "DELETE")
        self._add_delete_responses(operation)

    def get_request_body(self, path, method):
        """
        A request body is required by JSON:API for POST, PATCH, and DELETE methods.
        """
        serializer = self.get_serializer(path, method)
        if not isinstance(serializer, (serializers.BaseSerializer,)):
            return {}
        is_relationship = isinstance(self.view, views.RelationshipView)

        # DRF uses a $ref to the component schema definition, but this
        # doesn't work for JSON:API due to the different required fields based on
        # the method, so make those changes and inline another copy of the schema.
        # TODO: A future improvement could make this DRYer with multiple component schemas:
        #   A base schema for each viewset that has no required fields
        #   One subclassed from the base that requires some fields (`type` but not `id` for POST)
        #   Another subclassed from base with required type/id but no required attributes (PATCH)

        if is_relationship:
            item_schema = {"$ref": "#/components/schemas/ResourceIdentifierObject"}
        else:
            item_schema = self.map_serializer(serializer)
            if method == "POST":
                # 'type' and 'id' are both required for:
                # - all relationship operations
                # - regular PATCH or DELETE
                # Only 'type' is required for POST: system may assign the 'id'.
                item_schema["required"] = ["type"]

        if "properties" in item_schema and "attributes" in item_schema["properties"]:
            # No required attributes for PATCH
            if (
                method in ["PATCH", "PUT"]
                and "required" in item_schema["properties"]["attributes"]
            ):
                del item_schema["properties"]["attributes"]["required"]
            # No read_only fields for request.
            for name, schema in (
                item_schema["properties"]["attributes"]["properties"].copy().items()
            ):  # noqa E501
                if "readOnly" in schema:
                    del item_schema["properties"]["attributes"]["properties"][name]
        return {
            "content": {
                ct: {
                    "schema": {
                        "required": ["data"],
                        "properties": {"data": item_schema},
                    }
                }
                for ct in self.content_types
            }
        }

    def map_serializer(self, serializer):
        """
        Custom map_serializer that serializes the schema using the JSON:API spec.
        Non-attributes like related and identity fields, are move to 'relationships' and 'links'.
        """
        # TODO: remove attributes, etc. for relationshipView??
        required = []
        attributes = {}
        relationships = {}

        for field in serializer.fields.values():
            if isinstance(field, serializers.HyperlinkedIdentityField):
                # the 'url' is not an attribute but rather a self.link, so don't map it here.
                continue
            if isinstance(field, serializers.HiddenField):
                continue
            if isinstance(field, serializers.RelatedField):
                relationships[format_field_name(field.field_name)] = {
                    "$ref": "#/components/schemas/reltoone"
                }
                continue
            if isinstance(field, serializers.ManyRelatedField):
                relationships[format_field_name(field.field_name)] = {
                    "$ref": "#/components/schemas/reltomany"
                }
                continue

            if field.required:
                required.append(field.field_name)

            schema = self.map_field(field)
            if field.read_only:
                schema["readOnly"] = True
            if field.write_only:
                schema["writeOnly"] = True
            if field.allow_null:
                schema["nullable"] = True
            if field.default and field.default != empty:
                schema["default"] = field.default
            if field.help_text:
                # Ensure django gettext_lazy is rendered correctly
                schema["description"] = str(field.help_text)
            self.map_field_validators(field, schema)

            attributes[format_field_name(field.field_name)] = schema

        result = {
            "type": "object",
            "required": ["type", "id"],
            "additionalProperties": False,
            "properties": {
                "type": {"$ref": "#/components/schemas/type"},
                "id": {"$ref": "#/components/schemas/id"},
                "links": {
                    "type": "object",
                    "properties": {"self": {"$ref": "#/components/schemas/link"}},
                },
            },
        }
        if attributes:
            result["properties"]["attributes"] = {
                "type": "object",
                "properties": attributes,
            }
            if required:
                result["properties"]["attributes"]["required"] = required

        if relationships:
            result["properties"]["relationships"] = {
                "type": "object",
                "properties": relationships,
            }
        return result

    def _add_async_response(self, operation):
        """
        Add async response to operation
        """
        operation["responses"]["202"] = {
            "description": "Accepted for [asynchronous processing]"
            "(https://jsonapi.org/recommendations/#asynchronous-processing)",
            "content": {
                "application/vnd.api+json": {
                    "schema": {"$ref": "#/components/schemas/datum"}
                }
            },
        }

    def _failure_response(self, reason):
        """
        Return failure response reason as the description
        """
        return {
            "description": reason,
            "content": {
                "application/vnd.api+json": {
                    "schema": {"$ref": "#/components/schemas/failure"}
                }
            },
        }

    def _add_generic_failure_responses(self, operation):
        """
        Add generic failure response(s) to operation
        """
        for code, reason in [
            ("401", "not authorized"),
        ]:
            operation["responses"][code] = self._failure_response(reason)

    def _add_get_4xx_responses(self, operation):
        """
        Add generic 4xx GET responses to operation
        """
        self._add_generic_failure_responses(operation)
        for code, reason in [("404", "not found")]:
            operation["responses"][code] = self._failure_response(reason)

    def _add_post_4xx_responses(self, operation):
        """
        Add POST 4xx error responses to operation
        """
        self._add_generic_failure_responses(operation)
        for code, reason in [
            (
                "403",
                "[Forbidden](https://jsonapi.org/format/#crud-creating-responses-403)",
            ),
            (
                "404",
                "[Related resource does not exist]"
                "(https://jsonapi.org/format/#crud-creating-responses-404)",
            ),
            (
                "409",
                "[Conflict](https://jsonapi.org/format/#crud-creating-responses-409)",
            ),
        ]:
            operation["responses"][code] = self._failure_response(reason)

    def _add_patch_4xx_responses(self, operation):
        """
        Add PATCH 4xx error responses to operation
        """
        self._add_generic_failure_responses(operation)
        for code, reason in [
            (
                "403",
                "[Forbidden](https://jsonapi.org/format/#crud-updating-responses-403)",
            ),
            (
                "404",
                "[Related resource does not exist]"
                "(https://jsonapi.org/format/#crud-updating-responses-404)",
            ),
            (
                "409",
                "[Conflict]([Conflict]"
                "(https://jsonapi.org/format/#crud-updating-responses-409)",
            ),
        ]:
            operation["responses"][code] = self._failure_response(reason)

    def _add_delete_responses(self, operation):
        """
        Add generic DELETE responses to operation
        """
        # the 2xx statuses:
        operation["responses"] = {
            "200": {
                "description": "[OK](https://jsonapi.org/format/#crud-deleting-responses-200)",
                "content": {
                    "application/vnd.api+json": {
                        "schema": {"$ref": "#/components/schemas/onlymeta"}
                    }
                },
            }
        }
        self._add_async_response(operation)
        operation["responses"]["204"] = {
            "description": "[no content](https://jsonapi.org/format/#crud-deleting-responses-204)",
        }
        # the 4xx errors:
        self._add_generic_failure_responses(operation)
        for code, reason in [
            (
                "404",
                "[Resource does not exist]"
                "(https://jsonapi.org/format/#crud-deleting-responses-404)",
            ),
        ]:
            operation["responses"][code] = self._failure_response(reason)



from datetime import date, timedelta
from random import randint

from django.utils import timezone
from rest_framework.test import APITestCase

from example.factories import CommentFactory, EntryFactory
from example.models import Author, Blog, Comment, Entry, LabResults, ResearchProject


class PerformanceTestCase(APITestCase):
    def setUp(self):
        self.author = Author.objects.create(
            name="Super powerful superhero", email="i.am@lost.com"
        )
        self.blog = Blog.objects.create(name="Some Blog", tagline="It's a blog")
        self.other_blog = Blog.objects.create(
            name="Other blog", tagline="It's another blog"
        )
        self.first_entry = Entry.objects.create(
            blog=self.blog,
            headline="headline one",
            body_text="body_text two",
            pub_date=timezone.now(),
            mod_date=timezone.now(),
            n_comments=0,
            n_pingbacks=0,
            rating=3,
        )
        self.second_entry = Entry.objects.create(
            blog=self.blog,
            headline="headline two",
            body_text="body_text one",
            pub_date=timezone.now(),
            mod_date=timezone.now(),
            n_comments=0,
            n_pingbacks=0,
            rating=1,
        )
        self.comment = Comment.objects.create(entry=self.first_entry)
        CommentFactory.create_batch(50)
        EntryFactory.create_batch(50)

    def test_query_count_no_includes(self):
        """We expect a simple list view to issue only two queries.

        1. The number of results in the set (e.g. a COUNT query),
           only necessary because we're using PageNumberPagination
        2. The SELECT query for the set
        """
        with self.assertNumQueries(2):
            response = self.client.get("/comments?page[size]=25")
            self.assertEqual(len(response.data["results"]), 25)

    def test_query_count_include_author(self):
        """We expect a list view with an include have five queries:

        1. Primary resource COUNT query
        2. Primary resource SELECT
        3. Authors prefetched
        4. Author types prefetched
        5. Entries prefetched
        """
        with self.assertNumQueries(5):
            response = self.client.get("/comments?include=author&page[size]=25")
            self.assertEqual(len(response.data["results"]), 25)

    def test_query_select_related_entry(self):
        """We expect a list view with an include have two queries:

        1. Primary resource COUNT query
        2. Primary resource SELECT + SELECT RELATED writer(author) and bio
        """
        with self.assertNumQueries(2):
            response = self.client.get("/comments?include=writer&page[size]=25")
            self.assertEqual(len(response.data["results"]), 25)

    def test_query_prefetch_uses_included_resources(self):
        """We expect a list view with `included_resources` to have three queries:

        1. Primary resource COUNT query
        2. Primary resource SELECT
        3. Comments prefetched
        """
        with self.assertNumQueries(3):
            response = self.client.get(
                "/entries?fields[entries]=comments&page[size]=25"
            )
            self.assertEqual(len(response.data["results"]), 25)

    def test_query_prefetch_read_only(self):
        """We expect a read only list view with an include have five queries:

        1. Primary resource COUNT query
        2. Primary resource SELECT
        3. Authors prefetched
        4. Author types prefetched
        5. Entries prefetched
        """
        project = ResearchProject.objects.create(
            topic="Mars Mission", supervisor="Elon Musk"
        )

        LabResults.objects.bulk_create(
            [
                LabResults(
                    research_project=project,
                    date=date.today() + timedelta(days=i),
                    measurements=randint(0, 10000),
                    author=self.author,
                )
                for i in range(20)
            ]
        )

        with self.assertNumQueries(5):
            response = self.client.get("/lab-results?include=author&page[size]=25")
            self.assertEqual(len(response.data["results"]), 20)



from copy import deepcopy

import pytest
from django.urls import reverse
from rest_framework import status

from example import models, serializers, views

pytestmark = pytest.mark.django_db


class _PatchedModel:
    class JSONAPIMeta:
        resource_name = "resource_name_from_JSONAPIMeta"


def _check_resource_and_relationship_comment_type_match(django_client):
    entry_response = django_client.get(reverse("entry-list"))
    comment_response = django_client.get(reverse("comment-list"))

    comment_resource_type = comment_response.json().get("data")[0].get("type")
    comment_relationship_type = (
        entry_response.json()
        .get("data")[0]
        .get("relationships")
        .get("comments")
        .get("data")[0]
        .get("type")
    )

    assert (
        comment_resource_type == comment_relationship_type
    ), "The resource type seen in the relationships and head resource do not match"


def _check_relationship_and_included_comment_type_are_the_same(django_client, url):
    response = django_client.get(url + "?include=comments")
    data = response.json().get("data")[0]
    comment = response.json().get("included")[0]

    comment_relationship_type = (
        data.get("relationships").get("comments").get("data")[0].get("type")
    )
    comment_included_type = comment.get("type")

    assert (
        comment_relationship_type == comment_included_type
    ), "The resource type seen in the relationships and included do not match"


@pytest.mark.usefixtures("single_entry")
class TestModelResourceName:

    create_data = {
        "data": {
            "type": "resource_name_from_JSONAPIMeta",
            "id": None,
            "attributes": {
                "body": "example",
            },
            "relationships": {
                "entry": {"data": {"type": "resource_name_from_JSONAPIMeta", "id": 1}}
            },
        }
    }

    def test_model_resource_name_on_list(self, client):
        models.Comment.__bases__ += (_PatchedModel,)
        response = client.get(reverse("comment-list"))
        data = response.json()["data"][0]
        # name should be super-author instead of model name RenamedAuthor
        assert (
            data.get("type") == "resource_name_from_JSONAPIMeta"
        ), "resource_name from model incorrect on list"

    # Precedence tests
    def test_resource_name_precendence(self, client, monkeypatch):
        # default
        response = client.get(reverse("comment-list"))
        data = response.json()["data"][0]
        assert (
            data.get("type") == "comments"
        ), "resource_name from model incorrect on list"

        # model > default
        models.Comment.__bases__ += (_PatchedModel,)
        response = client.get(reverse("comment-list"))
        data = response.json()["data"][0]
        assert (
            data.get("type") == "resource_name_from_JSONAPIMeta"
        ), "resource_name from model incorrect on list"

        # serializer > model
        monkeypatch.setattr(
            serializers.CommentSerializer.Meta,
            "resource_name",
            "resource_name_from_serializer",
            False,
        )
        response = client.get(reverse("comment-list"))
        data = response.json()["data"][0]
        assert (
            data.get("type") == "resource_name_from_serializer"
        ), "resource_name from serializer incorrect on list"

        # view > serializer > model
        monkeypatch.setattr(
            views.CommentViewSet, "resource_name", "resource_name_from_view", False
        )
        response = client.get(reverse("comment-list"))
        data = response.json()["data"][0]
        assert (
            data.get("type") == "resource_name_from_view"
        ), "resource_name from view incorrect on list"

    def test_model_resource_name_create(self, client):
        models.Comment.__bases__ += (_PatchedModel,)
        models.Entry.__bases__ += (_PatchedModel,)
        response = client.post(reverse("comment-list"), self.create_data)

        assert response.status_code == status.HTTP_201_CREATED

    def test_serializer_resource_name_create(self, client, monkeypatch):
        monkeypatch.setattr(
            serializers.CommentSerializer.Meta,
            "resource_name",
            "renamed_comments",
            False,
        )
        monkeypatch.setattr(
            serializers.EntrySerializer.Meta, "resource_name", "renamed_entries", False
        )
        create_data = deepcopy(self.create_data)
        create_data["data"]["type"] = "renamed_comments"
        create_data["data"]["relationships"]["entry"]["data"][
            "type"
        ] = "renamed_entries"

        response = client.post(reverse("comment-list"), create_data)

        assert response.status_code == status.HTTP_201_CREATED

    def teardown_method(self, method):
        models.Comment.__bases__ = (models.Comment.__bases__[0],)
        models.Entry.__bases__ = (models.Entry.__bases__[0],)


@pytest.mark.usefixtures("single_entry")
class TestResourceNameConsistency:

    # Included rename tests
    def test_type_match_on_included_and_inline_base(self, client):
        _check_relationship_and_included_comment_type_are_the_same(
            client, reverse("entry-list")
        )

    def test_type_match_on_included_and_inline_with_JSONAPIMeta(self, client):
        models.Comment.__bases__ += (_PatchedModel,)

        _check_relationship_and_included_comment_type_are_the_same(
            client, reverse("entry-list")
        )

    def test_type_match_on_included_and_inline_with_serializer_resource_name(
        self, client
    ):
        serializers.CommentSerializer.Meta.resource_name = (
            "resource_name_from_serializer"
        )

        _check_relationship_and_included_comment_type_are_the_same(
            client, reverse("entry-list")
        )

    def test_type_match_on_included_and_inline_without_serializer_resource_name(
        self, client
    ):
        serializers.CommentSerializer.Meta.resource_name = None

        _check_relationship_and_included_comment_type_are_the_same(
            client, reverse("entry-list")
        )

    def test_type_match_on_included_and_inline_with_serializer_resource_name_and_JSONAPIMeta(
        self, client
    ):
        models.Comment.__bases__ += (_PatchedModel,)
        serializers.CommentSerializer.Meta.resource_name = (
            "resource_name_from_serializer"
        )

        _check_relationship_and_included_comment_type_are_the_same(
            client, reverse("entry-list")
        )

    # Relation rename tests
    def test_resource_and_relationship_type_match(self, client):
        _check_resource_and_relationship_comment_type_match(client)

    def test_resource_and_relationship_type_match_with_serializer_resource_name(
        self, client
    ):
        serializers.CommentSerializer.Meta.resource_name = (
            "resource_name_from_serializer"
        )

        _check_resource_and_relationship_comment_type_match(client)

    def test_resource_and_relationship_type_match_with_JSONAPIMeta(self, client):
        models.Comment.__bases__ += (_PatchedModel,)

        _check_resource_and_relationship_comment_type_match(client)

    def test_resource_and_relationship_type_match_with_serializer_resource_name_and_JSONAPIMeta(
        self, client
    ):
        models.Comment.__bases__ += (_PatchedModel,)
        serializers.CommentSerializer.Meta.resource_name = (
            "resource_name_from_serializer"
        )

        _check_resource_and_relationship_comment_type_match(client)

    def teardown_method(self, method):
        models.Comment.__bases__ = (models.Comment.__bases__[0],)
        try:
            delattr(serializers.CommentSerializer.Meta, "resource_name")
        except AttributeError:
            pass



# Generated by Django 3.0.3 on 2020-02-06 10:24

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("example", "0007_artproject_description"),
    ]

    operations = [
        migrations.CreateModel(
            name="LabResults",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("date", models.DateField()),
                ("measurements", models.TextField()),
                (
                    "research_project",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="lab_results",
                        to="example.ResearchProject",
                    ),
                ),
            ],
        ),
    ]



"""
This module provides the `json_api_settings` object that is used to access
Django REST framework JSON:API settings, checking for user settings first, then falling back to
the defaults.
"""

from django.conf import settings
from django.core.signals import setting_changed

JSON_API_SETTINGS_PREFIX = "JSON_API_"

DEFAULTS = {
    "FORMAT_FIELD_NAMES": False,
    "FORMAT_TYPES": False,
    "FORMAT_RELATED_LINKS": False,
    "PLURALIZE_TYPES": False,
    "UNIFORM_EXCEPTIONS": False,
}


class JSONAPISettings:
    """
    A settings object that allows JSON:API settings to be access as
    properties.
    """

    def __init__(self, user_settings=settings, defaults=DEFAULTS):
        self.defaults = defaults
        self.user_settings = user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid JSON:API setting: '%s'" % attr)

        value = getattr(
            self.user_settings, JSON_API_SETTINGS_PREFIX + attr, self.defaults[attr]
        )

        # Cache the result
        setattr(self, attr, value)
        return value


json_api_settings = JSONAPISettings()


def reload_json_api_settings(*args, **kwargs):
    django_setting = kwargs["setting"]
    setting = django_setting.replace(JSON_API_SETTINGS_PREFIX, "")
    value = kwargs["value"]
    if setting in DEFAULTS.keys():
        if value is not None:
            setattr(json_api_settings, setting, value)
        elif hasattr(json_api_settings, setting):
            delattr(json_api_settings, setting)


setting_changed.connect(reload_json_api_settings)



# largely based on DRF's test_openapi
import json

from django.test import RequestFactory, override_settings
from django.urls import re_path
from rest_framework.request import Request

from rest_framework_json_api.schemas.openapi import AutoSchema, SchemaGenerator

from example import views


def create_request(path):
    factory = RequestFactory()
    request = Request(factory.get(path))
    return request


def create_view_with_kw(view_cls, method, request, initkwargs):
    generator = SchemaGenerator()
    view = generator.create_view(view_cls.as_view(initkwargs), method, request)
    return view


def test_path_without_parameters(snapshot):
    path = "/authors/"
    method = "GET"

    view = create_view_with_kw(
        views.AuthorViewSet, method, create_request(path), {"get": "list"}
    )
    inspector = AutoSchema()
    inspector.view = view

    operation = inspector.get_operation(path, method)
    assert snapshot == json.dumps(operation, indent=2, sort_keys=True)


def test_path_with_id_parameter(snapshot):
    path = "/authors/{id}/"
    method = "GET"

    view = create_view_with_kw(
        views.AuthorViewSet, method, create_request(path), {"get": "retrieve"}
    )
    inspector = AutoSchema()
    inspector.view = view

    operation = inspector.get_operation(path, method)
    assert snapshot == json.dumps(operation, indent=2, sort_keys=True)


def test_post_request(snapshot):
    method = "POST"
    path = "/authors/"

    view = create_view_with_kw(
        views.AuthorViewSet, method, create_request(path), {"post": "create"}
    )
    inspector = AutoSchema()
    inspector.view = view

    operation = inspector.get_operation(path, method)
    assert snapshot == json.dumps(operation, indent=2, sort_keys=True)


def test_patch_request(snapshot):
    method = "PATCH"
    path = "/authors/{id}"

    view = create_view_with_kw(
        views.AuthorViewSet, method, create_request(path), {"patch": "update"}
    )
    inspector = AutoSchema()
    inspector.view = view

    operation = inspector.get_operation(path, method)
    assert snapshot == json.dumps(operation, indent=2, sort_keys=True)


def test_delete_request(snapshot):
    method = "DELETE"
    path = "/authors/{id}"

    view = create_view_with_kw(
        views.AuthorViewSet, method, create_request(path), {"delete": "delete"}
    )
    inspector = AutoSchema()
    inspector.view = view

    operation = inspector.get_operation(path, method)
    assert snapshot == json.dumps(operation, indent=2, sort_keys=True)


@override_settings(
    REST_FRAMEWORK={
        "DEFAULT_SCHEMA_CLASS": "rest_framework_json_api.schemas.openapi.AutoSchema"
    }
)
def test_schema_construction():
    """Construction of the top level dictionary."""
    patterns = [
        re_path("^authors/?$", views.AuthorViewSet.as_view({"get": "list"})),
    ]
    generator = SchemaGenerator(patterns=patterns)

    request = create_request("/")
    schema = generator.get_schema(request=request)

    assert "openapi" in schema
    assert "info" in schema
    assert "paths" in schema
    assert "components" in schema


def test_schema_related_serializers():
    """
    Confirm that paths are generated for related fields. For example:
        /authors/{pk}/{related_field>}
        /authors/{id}/comments/
        /authors/{id}/entries/
        /authors/{id}/first_entry/
    and confirm that the schema for the related field is properly rendered
    """
    generator = SchemaGenerator()
    request = create_request("/")
    schema = generator.get_schema(request=request)
    # make sure the path's relationship and related {related_field}'s got expanded
    assert "/authors/{id}/relationships/{related_field}" in schema["paths"]
    assert "/authors/{id}/comments/" in schema["paths"]
    assert "/authors/{id}/entries/" in schema["paths"]
    assert "/authors/{id}/first_entry/" in schema["paths"]
    first_get = schema["paths"]["/authors/{id}/first_entry/"]["get"]["responses"]["200"]
    first_schema = first_get["content"]["application/vnd.api+json"]["schema"]
    first_props = first_schema["properties"]["data"]
    assert "$ref" in first_props
    assert first_props["$ref"] == "#/components/schemas/Entry"



"""
Parsers
"""
from rest_framework import parsers
from rest_framework.exceptions import ParseError

from rest_framework_json_api import exceptions, renderers, serializers
from rest_framework_json_api.utils import get_resource_name, undo_format_field_names


class JSONParser(parsers.JSONParser):
    """
    Similar to `JSONRenderer`, the `JSONParser` you may override the following methods if you
    need highly custom parsing control.

    A JSON:API client will send a payload that looks like this:

    .. code:: json


        {
            "data": {
                "type": "identities",
                "id": 1,
                "attributes": {
                    "first_name": "John",
                    "last_name": "Coltrane"
                }
            }
        }

    We extract the attributes so that DRF serializers can work as normal.
    """

    media_type = "application/vnd.api+json"
    renderer_class = renderers.JSONRenderer

    @staticmethod
    def parse_attributes(data):
        attributes = data.get("attributes") or dict()
        return undo_format_field_names(attributes)

    @staticmethod
    def parse_relationships(data):
        relationships = data.get("relationships") or dict()
        relationships = undo_format_field_names(relationships)

        # Parse the relationships
        parsed_relationships = dict()
        for field_name, field_data in relationships.items():
            field_data = field_data.get("data")
            if isinstance(field_data, dict) or field_data is None:
                parsed_relationships[field_name] = field_data
            elif isinstance(field_data, list):
                parsed_relationships[field_name] = list(
                    relation for relation in field_data
                )
        return parsed_relationships

    @staticmethod
    def parse_metadata(result):
        """
        Returns a dictionary which will be merged into parsed data of the request. By default,
        it reads the `meta` content in the request body and returns it in a dictionary with
        a `_meta` top level key.
        """
        metadata = result.get("meta")
        if metadata:
            return {"_meta": metadata}
        else:
            return {}

    def parse(self, stream, media_type=None, parser_context=None):
        """
        Parses the incoming bytestream as JSON and returns the resulting data
        """
        result = super().parse(
            stream, media_type=media_type, parser_context=parser_context
        )

        if not isinstance(result, dict) or "data" not in result:
            raise ParseError("Received document does not contain primary data")

        data = result.get("data")
        parser_context = parser_context or {}
        view = parser_context.get("view")

        from rest_framework_json_api.views import RelationshipView

        if isinstance(view, RelationshipView):
            # We skip parsing the object as JSON:API Resource Identifier Object and not a regular
            # Resource Object
            if isinstance(data, list):
                for resource_identifier_object in data:
                    if not (
                        resource_identifier_object.get("id")
                        and resource_identifier_object.get("type")
                    ):
                        raise ParseError(
                            "Received data contains one or more malformed JSON:API "
                            "Resource Identifier Object(s)"
                        )
            elif not (data.get("id") and data.get("type")):
                raise ParseError(
                    "Received data is not a valid JSON:API Resource Identifier Object"
                )

            return data

        request = parser_context.get("request")
        method = request and request.method

        # Sanity check
        if not isinstance(data, dict):
            raise ParseError(
                "Received data is not a valid JSON:API Resource Identifier Object"
            )

        # Check for inconsistencies
        if method in ("PUT", "POST", "PATCH"):
            resource_name = get_resource_name(
                parser_context, expand_polymorphic_types=True
            )
            if isinstance(resource_name, str):
                if data.get("type") != resource_name:
                    raise exceptions.Conflict(
                        "The resource object's type ({data_type}) is not the type that "
                        "constitute the collection represented by the endpoint "
                        "({resource_type}).".format(
                            data_type=data.get("type"), resource_type=resource_name
                        )
                    )
            else:
                if data.get("type") not in resource_name:
                    raise exceptions.Conflict(
                        "The resource object's type ({data_type}) is not the type that "
                        "constitute the collection represented by the endpoint "
                        "(one of [{resource_types}]).".format(
                            data_type=data.get("type"),
                            resource_types=", ".join(resource_name),
                        )
                    )
        if not data.get("id") and method in ("PATCH", "PUT"):
            raise ParseError(
                "The resource identifier object must contain an 'id' member"
            )

        if method in ("PATCH", "PUT"):
            lookup_url_kwarg = getattr(view, "lookup_url_kwarg", None) or getattr(
                view, "lookup_field", None
            )
            if lookup_url_kwarg and str(data.get("id")) != str(
                view.kwargs[lookup_url_kwarg]
            ):
                raise exceptions.Conflict(
                    "The resource object's id ({data_id}) does not match url's "
                    "lookup id ({url_id})".format(
                        data_id=data.get("id"), url_id=view.kwargs[lookup_url_kwarg]
                    )
                )

        # Construct the return data
        serializer_class = getattr(view, "serializer_class", None)
        parsed_data = {"id": data.get("id")} if "id" in data else {}
        # TODO remove in next major version 5.0.0 see serializers.ReservedFieldNamesMixin
        if serializer_class is not None:
            if issubclass(serializer_class, serializers.PolymorphicModelSerializer):
                parsed_data["type"] = data.get("type")
        parsed_data.update(self.parse_attributes(data))
        parsed_data.update(self.parse_relationships(data))
        parsed_data.update(self.parse_metadata(result))
        return parsed_data



from collections import OrderedDict

from rest_framework.request import Request
from rest_framework.test import APIRequestFactory
from rest_framework.utils.urls import replace_query_param

from rest_framework_json_api import pagination

factory = APIRequestFactory()


class TestLimitOffset:
    """
    Unit tests for `pagination.JsonApiLimitOffsetPagination`.
    """

    def setup(self):
        class ExamplePagination(pagination.JsonApiLimitOffsetPagination):
            default_limit = 10
            max_limit = 15

        self.pagination = ExamplePagination()
        self.queryset = range(1, 101)
        self.base_url = "http://testserver/"

    def paginate_queryset(self, request):
        return list(self.pagination.paginate_queryset(self.queryset, request))

    def get_paginated_content(self, queryset):
        response = self.pagination.get_paginated_response(queryset)
        return response.data

    def get_test_request(self, arguments):
        return Request(factory.get("/", arguments))

    def test_valid_offset_limit(self):
        """
        Basic test, assumes offset and limit are given.
        """
        offset = 10
        limit = 5
        count = len(self.queryset)
        last_offset = (count // limit) * limit
        next_offset = 15
        prev_offset = 5

        request = self.get_test_request(
            {
                self.pagination.limit_query_param: limit,
                self.pagination.offset_query_param: offset,
            }
        )
        base_url = replace_query_param(
            self.base_url, self.pagination.limit_query_param, limit
        )
        last_url = replace_query_param(
            base_url, self.pagination.offset_query_param, last_offset
        )
        first_url = base_url
        next_url = replace_query_param(
            base_url, self.pagination.offset_query_param, next_offset
        )
        prev_url = replace_query_param(
            base_url, self.pagination.offset_query_param, prev_offset
        )
        queryset = self.paginate_queryset(request)
        content = self.get_paginated_content(queryset)
        next_offset = offset + limit

        expected_content = {
            "results": list(range(offset + 1, next_offset + 1)),
            "links": OrderedDict(
                [
                    ("first", first_url),
                    ("last", last_url),
                    ("next", next_url),
                    ("prev", prev_url),
                ]
            ),
            "meta": {
                "pagination": OrderedDict(
                    [
                        ("count", count),
                        ("limit", limit),
                        ("offset", offset),
                    ]
                )
            },
        }

        assert queryset == list(range(offset + 1, next_offset + 1))
        assert content == expected_content



import os

SITE_ID = 1
DEBUG = True

MEDIA_ROOT = os.path.normcase(os.path.dirname(os.path.abspath(__file__)))
MEDIA_URL = "/media/"
USE_TZ = False

DATABASE_ENGINE = "sqlite3"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "drf_example",
    }
}

INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    "django.contrib.sessions",
    "django.contrib.auth",
    "rest_framework_json_api",
    "rest_framework",
    "polymorphic",
    "example",
    "debug_toolbar",
    "django_filters",
    "tests",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [
            # insert your TEMPLATE_DIRS here
        ],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                # Insert your TEMPLATE_CONTEXT_PROCESSORS here or use this
                # list if you haven't customized them:
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

STATIC_URL = "/static/"

ROOT_URLCONF = "example.urls"

SECRET_KEY = "abc123"

PASSWORD_HASHERS = ("django.contrib.auth.hashers.UnsaltedMD5PasswordHasher",)

MIDDLEWARE = ("debug_toolbar.middleware.DebugToolbarMiddleware",)

INTERNAL_IPS = ("127.0.0.1",)

JSON_API_FORMAT_FIELD_NAMES = "camelize"
JSON_API_FORMAT_TYPES = "camelize"
REST_FRAMEWORK = {
    "PAGE_SIZE": 5,
    "EXCEPTION_HANDLER": "rest_framework_json_api.exceptions.exception_handler",
    "DEFAULT_PAGINATION_CLASS": "rest_framework_json_api.pagination.JsonApiPageNumberPagination",
    "DEFAULT_PARSER_CLASSES": (
        "rest_framework_json_api.parsers.JSONParser",
        "rest_framework.parsers.FormParser",
        "rest_framework.parsers.MultiPartParser",
    ),
    "DEFAULT_RENDERER_CLASSES": (
        "rest_framework_json_api.renderers.JSONRenderer",
        # If you're performance testing, you will want to use the browseable API
        # without forms, as the forms can generate their own queries.
        # If performance testing, enable:
        # 'example.utils.BrowsableAPIRendererWithoutForms',
        # Otherwise, to play around with the browseable API, enable:
        "rest_framework_json_api.renderers.BrowsableAPIRenderer",
    ),
    "DEFAULT_METADATA_CLASS": "rest_framework_json_api.metadata.JSONAPIMetadata",
    "DEFAULT_SCHEMA_CLASS": "rest_framework_json_api.schemas.openapi.AutoSchema",
    "DEFAULT_FILTER_BACKENDS": (
        "rest_framework_json_api.filters.OrderingFilter",
        "rest_framework_json_api.django_filters.DjangoFilterBackend",
        "rest_framework.filters.SearchFilter",
    ),
    "SEARCH_PARAM": "filter[search]",
    "TEST_REQUEST_RENDERER_CLASSES": (
        "rest_framework_json_api.renderers.JSONRenderer",
    ),
    "TEST_REQUEST_DEFAULT_FORMAT": "vnd.api+json",
}



import pytest

from example.factories import BlogFactory
from example.models import Blog

pytestmark = pytest.mark.django_db


def test_factory_instance(blog_factory):

    assert blog_factory == BlogFactory


def test_model_instance(blog):

    assert isinstance(blog, Blog)


def test_multiple_blog(blog_factory):
    another_blog = blog_factory(name="Cool Blog")
    new_blog = blog_factory(name="Awesome Blog")

    assert another_blog.name == "Cool Blog"
    assert new_blog.name == "Awesome Blog"


def test_factories_with_relations(author_factory, entry_factory):

    author = author_factory(name="Joel Spolsky")
    entry = entry_factory(
        headline=(
            "The Absolute Minimum Every Software Developer"
            "Absolutely, Positively Must Know About Unicode "
            "and Character Sets (No Excuses!)"
        ),
        blog__name="Joel on Software",
        authors=(author,),
    )

    assert entry.blog.name == "Joel on Software"
    assert entry.headline == (
        "The Absolute Minimum Every Software Developer"
        "Absolutely, Positively Must Know About Unicode "
        "and Character Sets (No Excuses!)"
    )
    assert entry.authors.all().count() == 1
    assert entry.authors.all()[0].name == "Joel Spolsky"



# Generated by Django 2.1.1 on 2018-09-22 15:08

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("example", "0004_auto_20171011_0631"),
    ]

    operations = [
        migrations.CreateModel(
            name="ProjectType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=50)),
            ],
            options={
                "ordering": ("id",),
            },
        ),
        migrations.AlterModelOptions(
            name="artproject",
            options={"base_manager_name": "objects"},
        ),
        migrations.AlterModelOptions(
            name="project",
            options={"base_manager_name": "objects"},
        ),
        migrations.AlterModelOptions(
            name="researchproject",
            options={"base_manager_name": "objects"},
        ),
        migrations.AddField(
            model_name="project",
            name="project_type",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="example.ProjectType",
            ),
        ),
    ]



from datetime import datetime

from rest_framework import fields as drf_fields
from rest_framework import serializers as drf_serilazers

from rest_framework_json_api import relations, serializers

from example.models import (
    ArtProject,
    Author,
    AuthorBio,
    AuthorBioMetadata,
    AuthorType,
    Blog,
    Comment,
    Company,
    Entry,
    LabResults,
    Project,
    ProjectType,
    ResearchProject,
    TaggedItem,
)


class TaggedItemSerializer(serializers.ModelSerializer):
    class Meta:
        model = TaggedItem
        fields = ("tag",)


class TaggedItemDRFSerializer(drf_serilazers.ModelSerializer):
    """
    DRF default serializer to test default DRF functionalities
    """

    class Meta:
        model = TaggedItem
        fields = ("tag",)


class BlogSerializer(serializers.ModelSerializer):
    copyright = serializers.SerializerMethodField()
    tags = relations.ResourceRelatedField(many=True, read_only=True)

    included_serializers = {
        "tags": "example.serializers.TaggedItemSerializer",
    }

    def get_copyright(self, resource):
        return datetime.now().year

    def get_root_meta(self, resource, many):
        return {"api_docs": "/docs/api/blogs"}

    class Meta:
        model = Blog
        fields = ("name", "url", "tags")
        read_only_fields = ("tags",)
        meta_fields = ("copyright",)


class BlogDRFSerializer(drf_serilazers.ModelSerializer):
    """
    DRF default serializer to test default DRF functionalities
    """

    copyright = serializers.SerializerMethodField()
    tags = TaggedItemDRFSerializer(many=True, read_only=True)

    def get_copyright(self, resource):
        return datetime.now().year

    def get_root_meta(self, resource, many):
        return {"api_docs": "/docs/api/blogs"}

    class Meta:
        model = Blog
        fields = ("name", "url", "tags", "copyright")
        read_only_fields = ("tags",)
        meta_fields = ("copyright",)


class EntrySerializer(serializers.ModelSerializer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # to make testing more concise we'll only output the
        # `featured` field when it's requested via `include`
        request = kwargs.get("context", {}).get("request")
        if request and "featured" not in request.query_params.get("include", []):
            self.fields.pop("featured", None)

    included_serializers = {
        "authors": "example.serializers.AuthorSerializer",
        "comments": "example.serializers.CommentSerializer",
        "featured": "example.serializers.EntrySerializer",
        "suggested": "example.serializers.EntrySerializer",
        "tags": "example.serializers.TaggedItemSerializer",
    }

    body_format = serializers.SerializerMethodField()
    # single related from model
    blog_hyperlinked = relations.HyperlinkedRelatedField(
        related_link_view_name="entry-blog",
        related_link_url_kwarg="entry_pk",
        self_link_view_name="entry-relationships",
        read_only=True,
        source="blog",
    )
    # many related from model
    comments = relations.ResourceRelatedField(many=True, read_only=True)
    # many related hyperlinked from model
    comments_hyperlinked = relations.HyperlinkedRelatedField(
        related_link_view_name="entry-comments",
        related_link_url_kwarg="entry_pk",
        self_link_view_name="entry-relationships",
        many=True,
        read_only=True,
        source="comments",
    )
    # many related from serializer
    suggested = relations.SerializerMethodResourceRelatedField(
        related_link_view_name="entry-suggested",
        related_link_url_kwarg="entry_pk",
        self_link_view_name="entry-relationships",
        model=Entry,
        many=True,
    )
    # many related hyperlinked from serializer
    suggested_hyperlinked = relations.SerializerMethodHyperlinkedRelatedField(
        related_link_view_name="entry-suggested",
        related_link_url_kwarg="entry_pk",
        self_link_view_name="entry-relationships",
        model=Entry,
        many=True,
    )
    # single related from serializer
    featured = relations.SerializerMethodResourceRelatedField(model=Entry)
    # single related hyperlinked from serializer
    featured_hyperlinked = relations.SerializerMethodHyperlinkedRelatedField(
        related_link_view_name="entry-featured",
        related_link_url_kwarg="entry_pk",
        self_link_view_name="entry-relationships",
        model=Entry,
        read_only=True,
    )
    tags = relations.ResourceRelatedField(many=True, read_only=True)

    def get_suggested(self, obj):
        return Entry.objects.exclude(pk=obj.pk)

    def get_featured(self, obj):
        return Entry.objects.exclude(pk=obj.pk).first()

    def get_body_format(self, obj):
        return "text"

    class Meta:
        model = Entry
        fields = (
            "blog",
            "blog_hyperlinked",
            "headline",
            "body_text",
            "pub_date",
            "mod_date",
            "authors",
            "comments",
            "comments_hyperlinked",
            "featured",
            "suggested",
            "suggested_hyperlinked",
            "tags",
            "featured_hyperlinked",
        )
        read_only_fields = ("tags",)
        meta_fields = ("body_format",)

    class JSONAPIMeta:
        included_resources = ["comments"]


class EntryDRFSerializers(drf_serilazers.ModelSerializer):

    tags = TaggedItemDRFSerializer(many=True, read_only=True)
    url = drf_serilazers.HyperlinkedIdentityField(
        view_name="drf-entry-blog-detail",
        lookup_url_kwarg="entry_pk",
        read_only=True,
    )

    class Meta:
        model = Entry
        fields = (
            "tags",
            "url",
        )
        read_only_fields = ("tags",)


class AuthorTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = AuthorType
        fields = ("name",)


class AuthorBioSerializer(serializers.ModelSerializer):
    class Meta:
        model = AuthorBio
        fields = ("author", "body", "metadata")

    included_serializers = {
        "metadata": "example.serializers.AuthorBioMetadataSerializer",
    }


class AuthorBioMetadataSerializer(serializers.ModelSerializer):
    class Meta:
        model = AuthorBioMetadata
        fields = ("body",)


class AuthorSerializer(serializers.ModelSerializer):
    bio = relations.ResourceRelatedField(
        related_link_view_name="author-related",
        self_link_view_name="author-relationships",
        queryset=AuthorBio.objects,
    )
    entries = relations.ResourceRelatedField(
        related_link_view_name="author-related",
        self_link_view_name="author-relationships",
        queryset=Entry.objects,
        many=True,
    )
    first_entry = relations.SerializerMethodResourceRelatedField(
        related_link_view_name="author-related",
        self_link_view_name="author-relationships",
        model=Entry,
    )
    comments = relations.HyperlinkedRelatedField(
        related_link_view_name="author-related",
        self_link_view_name="author-relationships",
        queryset=Comment.objects,
        many=True,
    )
    secrets = serializers.HiddenField(default="Shhhh!")
    defaults = serializers.CharField(
        default="default",
        max_length=20,
        min_length=3,
        write_only=True,
        help_text="help for defaults",
    )
    initials = serializers.SerializerMethodField()
    included_serializers = {"bio": AuthorBioSerializer, "type": AuthorTypeSerializer}
    related_serializers = {
        "bio": "example.serializers.AuthorBioSerializer",
        "type": "example.serializers.AuthorTypeSerializer",
        "comments": "example.serializers.CommentSerializer",
        "entries": "example.serializers.EntrySerializer",
        "first_entry": "example.serializers.EntrySerializer",
    }

    class Meta:
        model = Author
        fields = (
            "name",
            "email",
            "bio",
            "entries",
            "comments",
            "first_entry",
            "type",
            "secrets",
            "defaults",
            "initials",
        )
        meta_fields = ("initials",)

    def get_first_entry(self, obj):
        return obj.entries.first()

    def get_initials(self, obj):
        return "".join([word[0] for word in obj.name.split(" ")])


class AuthorListSerializer(AuthorSerializer):
    pass


class AuthorDetailSerializer(AuthorSerializer):
    pass


class WriterSerializer(serializers.ModelSerializer):
    included_serializers = {"bio": AuthorBioSerializer}

    class Meta:
        model = Author
        fields = ("name", "email", "bio")
        resource_name = "writers"


class CommentSerializer(serializers.ModelSerializer):
    # testing remapping of related name
    writer = relations.ResourceRelatedField(source="author", read_only=True)
    modified_days_ago = serializers.SerializerMethodField()

    included_serializers = {
        "entry": EntrySerializer,
        "author": AuthorSerializer,
        "writer": WriterSerializer,
    }

    class Meta:
        model = Comment
        exclude = (
            "created_at",
            "modified_at",
        )
        # fields = ('entry', 'body', 'author',)
        meta_fields = ("modified_days_ago",)

    class JSONAPIMeta:
        included_resources = ("writer",)

    def get_modified_days_ago(self, obj):
        return (datetime.now() - obj.modified_at).days


class ProjectTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = ProjectType
        fields = (
            "name",
            "url",
        )


class BaseProjectSerializer(serializers.ModelSerializer):
    included_serializers = {
        "project_type": ProjectTypeSerializer,
    }


class ArtProjectSerializer(BaseProjectSerializer):
    class Meta:
        model = ArtProject
        exclude = ("polymorphic_ctype",)


class ResearchProjectSerializer(BaseProjectSerializer):
    # testing exclusive related field on inherited polymorphic model
    lab_results = relations.ResourceRelatedField(many=True, read_only=True)

    class Meta:
        model = ResearchProject
        exclude = ("polymorphic_ctype",)


class LabResultsSerializer(serializers.ModelSerializer):
    included_serializers = {"author": AuthorSerializer}

    class Meta:
        model = LabResults
        fields = ("date", "measurements", "author")


class ProjectSerializer(serializers.PolymorphicModelSerializer):
    included_serializers = {
        "project_type": ProjectTypeSerializer,
    }
    polymorphic_serializers = [ArtProjectSerializer, ResearchProjectSerializer]

    class Meta:
        model = Project
        exclude = ("polymorphic_ctype",)


class CurrentProjectRelatedField(relations.PolymorphicResourceRelatedField):
    def get_attribute(self, instance):
        obj = super().get_attribute(instance)

        is_art = self.field_name == "current_art_project" and isinstance(
            obj, ArtProject
        )
        is_res = self.field_name == "current_research_project" and isinstance(
            obj, ResearchProject
        )

        if is_art or is_res:
            return obj

        raise drf_fields.SkipField()


class CompanySerializer(serializers.ModelSerializer):
    current_project = relations.PolymorphicResourceRelatedField(
        ProjectSerializer, queryset=Project.objects.all()
    )
    current_art_project = CurrentProjectRelatedField(
        ProjectSerializer, source="current_project", read_only=True
    )
    current_research_project = CurrentProjectRelatedField(
        ProjectSerializer, source="current_project", read_only=True
    )
    future_projects = relations.PolymorphicResourceRelatedField(
        ProjectSerializer, queryset=Project.objects.all(), many=True
    )

    included_serializers = {
        "current_project": ProjectSerializer,
        "future_projects": ProjectSerializer,
        "current_art_project": ProjectSerializer,
        "current_research_project": ProjectSerializer,
    }

    class Meta:
        model = Company
        fields = "__all__"



import pytest
from django.urls import re_path
from rest_framework import status
from rest_framework.fields import SkipField
from rest_framework.routers import SimpleRouter
from rest_framework.serializers import Serializer

from rest_framework_json_api.exceptions import Conflict
from rest_framework_json_api.relations import (
    HyperlinkedRelatedField,
    SerializerMethodHyperlinkedRelatedField,
)
from rest_framework_json_api.utils import format_link_segment
from rest_framework_json_api.views import RelationshipView
from tests.models import BasicModel
from tests.serializers import (
    ForeignKeySourceSerializer,
    ManyToManySourceReadOnlySerializer,
    ManyToManySourceSerializer,
)
from tests.views import BasicModelViewSet


@pytest.mark.django_db
class TestResourceRelatedField:
    @pytest.mark.parametrize(
        "format_type,pluralize_type,resource_type",
        [
            (False, False, "ForeignKeyTarget"),
            (False, True, "ForeignKeyTargets"),
            ("dasherize", False, "foreign-key-target"),
            ("dasherize", True, "foreign-key-targets"),
        ],
    )
    def test_serialize(
        self, format_type, pluralize_type, resource_type, foreign_key_target, settings
    ):
        settings.JSON_API_FORMAT_TYPES = format_type
        settings.JSON_API_PLURALIZE_TYPES = pluralize_type

        serializer = ForeignKeySourceSerializer(instance={"target": foreign_key_target})
        expected = {
            "type": resource_type,
            "id": str(foreign_key_target.pk),
        }

        assert serializer.data["target"] == expected

    @pytest.mark.parametrize(
        "format_type,pluralize_type,resource_type",
        [
            (False, False, "ForeignKeyTarget"),
            (False, True, "ForeignKeyTargets"),
            ("dasherize", False, "foreign-key-target"),
            ("dasherize", True, "foreign-key-targets"),
        ],
    )
    def test_deserialize(
        self, format_type, pluralize_type, resource_type, foreign_key_target, settings
    ):
        settings.JSON_API_FORMAT_TYPES = format_type
        settings.JSON_API_PLURALIZE_TYPES = pluralize_type

        serializer = ForeignKeySourceSerializer(
            data={"target": {"type": resource_type, "id": str(foreign_key_target.pk)}}
        )

        assert serializer.is_valid()
        assert serializer.validated_data["target"] == foreign_key_target

    @pytest.mark.parametrize(
        "format_type,pluralize_type,resource_type",
        [
            (False, False, "ForeignKeyTargets"),
            (False, False, "Invalid"),
            (False, False, "foreign-key-target"),
            (False, True, "ForeignKeyTarget"),
            ("dasherize", False, "ForeignKeyTarget"),
            ("dasherize", True, "ForeignKeyTargets"),
        ],
    )
    def test_validation_fails_on_invalid_type(
        self, format_type, pluralize_type, resource_type, foreign_key_target, settings
    ):
        settings.JSON_API_FORMAT_TYPES = format_type
        settings.JSON_API_PLURALIZE_TYPES = pluralize_type

        with pytest.raises(Conflict) as e:
            serializer = ForeignKeySourceSerializer(
                data={
                    "target": {"type": resource_type, "id": str(foreign_key_target.pk)}
                }
            )
            serializer.is_valid()
        assert e.value.status_code == status.HTTP_409_CONFLICT

    @pytest.mark.parametrize(
        "format_type,pluralize_type,resource_type",
        [
            (False, False, "ManyToManyTarget"),
            (False, True, "ManyToManyTargets"),
            ("dasherize", False, "many-to-many-target"),
            ("dasherize", True, "many-to-many-targets"),
        ],
    )
    def test_serialize_many_to_many_relation(
        self,
        format_type,
        pluralize_type,
        resource_type,
        many_to_many_source,
        many_to_many_targets,
        settings,
    ):
        settings.JSON_API_FORMAT_TYPES = format_type
        settings.JSON_API_PLURALIZE_TYPES = pluralize_type

        serializer = ManyToManySourceSerializer(instance=many_to_many_source)
        expected = [
            {"type": resource_type, "id": str(target.pk)}
            for target in many_to_many_targets
        ]
        assert serializer.data["targets"] == expected

    @pytest.mark.parametrize(
        "format_type,pluralize_type,resource_type",
        [
            (False, False, "ManyToManyTarget"),
            (False, True, "ManyToManyTargets"),
            ("dasherize", False, "many-to-many-target"),
            ("dasherize", True, "many-to-many-targets"),
        ],
    )
    @pytest.mark.parametrize(
        "serializer_class",
        [ManyToManySourceSerializer, ManyToManySourceReadOnlySerializer],
    )
    def test_deserialize_many_to_many_relation(
        self,
        format_type,
        pluralize_type,
        resource_type,
        serializer_class,
        many_to_many_targets,
        settings,
    ):
        settings.JSON_API_FORMAT_TYPES = format_type
        settings.JSON_API_PLURALIZE_TYPES = pluralize_type

        targets = [
            {"type": resource_type, "id": target.pk} for target in many_to_many_targets
        ]
        serializer = ManyToManySourceSerializer(data={"targets": targets})
        assert serializer.is_valid()
        assert serializer.validated_data["targets"] == many_to_many_targets

    @pytest.mark.parametrize(
        "resource_identifier,error",
        [
            (
                {"type": "ForeignKeyTarget"},
                "Invalid resource identifier object: missing 'id' attribute",
            ),
            (
                {"id": "1234"},
                "Invalid resource identifier object: missing 'type' attribute",
            ),
        ],
    )
    def test_invalid_resource_id_object(self, resource_identifier, error):
        serializer = ForeignKeySourceSerializer(data={"target": resource_identifier})
        assert not serializer.is_valid()
        assert serializer.errors == {"target": [error]}


class TestHyperlinkedRelatedField:
    @pytest.fixture
    def instance(self):
        # dummy instance
        return object()

    @pytest.fixture
    def serializer(self):
        class HyperlinkedRelatedFieldSerializer(Serializer):
            single = HyperlinkedRelatedField(
                self_link_view_name="basic-model-relationships",
                related_link_view_name="basic-model-related",
                read_only=True,
            )
            many = HyperlinkedRelatedField(
                self_link_view_name="basic-model-relationships",
                related_link_view_name="basic-model-related",
                read_only=True,
                many=True,
            )
            single_serializer_method = SerializerMethodHyperlinkedRelatedField(
                self_link_view_name="basic-model-relationships",
                related_link_view_name="basic-model-related",
                read_only=True,
            )
            many_serializer_method = SerializerMethodHyperlinkedRelatedField(
                self_link_view_name="basic-model-relationships",
                related_link_view_name="basic-model-related",
                read_only=True,
                many=True,
            )

            def get_single_serializer_method(self, obj):  # pragma: no cover
                raise NotImplementedError

            def get_many_serializer_method(self, obj):  # pragma: no cover
                raise NotImplementedError

        return HyperlinkedRelatedFieldSerializer()

    @pytest.fixture(
        params=["single", "many", "single_serializer_method", "many_serializer_method"]
    )
    def field(self, serializer, request):
        field = serializer.fields[request.param]
        field.field_name = request.param
        return field

    def test_get_attribute(self, model, field):
        with pytest.raises(SkipField):
            field.get_attribute(model)

    def test_to_representation(self, model, field):
        with pytest.raises(NotImplementedError):
            field.to_representation(model)

    @pytest.mark.urls(__name__)
    @pytest.mark.parametrize(
        "format_related_links",
        [
            False,
            "dasherize",
            "camelize",
            "capitalize",
            "underscore",
        ],
    )
    def test_get_links(
        self,
        format_related_links,
        field,
        settings,
        model,
    ):
        settings.JSON_API_FORMAT_RELATED_LINKS = format_related_links

        link_segment = format_link_segment(field.field_name)

        expected = {
            "self": f"/basic_models/{model.pk}/relationships/{link_segment}/",
            "related": f"/basic_models/{model.pk}/{link_segment}/",
        }

        if hasattr(field, "child_relation"):
            # many case
            field = field.child_relation

        actual = field.get_links(model)
        assert expected == actual


# Routing setup


class BasicModelRelationshipView(RelationshipView):
    queryset = BasicModel.objects


router = SimpleRouter()
router.register(r"basic_models", BasicModelViewSet, basename="basic-model")

urlpatterns = [
    re_path(
        r"^basic_models/(?P<pk>[^/.]+)/(?P<related_field>[^/.]+)/$",
        BasicModelViewSet.as_view({"get": "retrieve_related"}),
        name="basic-model-related",
    ),
    re_path(
        r"^basic_models/(?P<pk>[^/.]+)/relationships/(?P<related_field>[^/.]+)/$",
        BasicModelRelationshipView.as_view(),
        name="basic-model-relationships",
    ),
]

urlpatterns += router.urls



from django.conf import settings
from django.conf.urls import include, url
from django.urls import path
from django.views.generic import TemplateView
from rest_framework import routers
from rest_framework.schemas import get_schema_view

from rest_framework_json_api.schemas.openapi import SchemaGenerator

from example.views import (
    AuthorRelationshipView,
    AuthorViewSet,
    BlogRelationshipView,
    BlogViewSet,
    CommentRelationshipView,
    CommentViewSet,
    CompanyViewset,
    EntryRelationshipView,
    EntryViewSet,
    LabResultViewSet,
    NonPaginatedEntryViewSet,
    ProjectTypeViewset,
    ProjectViewset,
)

router = routers.DefaultRouter(trailing_slash=False)

router.register(r"blogs", BlogViewSet)
router.register(r"entries", EntryViewSet)
router.register(r"nopage-entries", NonPaginatedEntryViewSet, "nopage-entry")
router.register(r"authors", AuthorViewSet)
router.register(r"comments", CommentViewSet)
router.register(r"companies", CompanyViewset)
router.register(r"projects", ProjectViewset)
router.register(r"project-types", ProjectTypeViewset)
router.register(r"lab-results", LabResultViewSet)

urlpatterns = [
    url(r"^", include(router.urls)),
    url(
        r"^entries/(?P<entry_pk>[^/.]+)/suggested/$",
        EntryViewSet.as_view({"get": "list"}),
        name="entry-suggested",
    ),
    url(
        r"entries/(?P<entry_pk>[^/.]+)/blog$",
        BlogViewSet.as_view({"get": "retrieve"}),
        name="entry-blog",
    ),
    url(
        r"entries/(?P<entry_pk>[^/.]+)/comments$",
        CommentViewSet.as_view({"get": "list"}),
        name="entry-comments",
    ),
    url(
        r"entries/(?P<entry_pk>[^/.]+)/authors$",
        AuthorViewSet.as_view({"get": "list"}),
        name="entry-authors",
    ),
    url(
        r"entries/(?P<entry_pk>[^/.]+)/featured$",
        EntryViewSet.as_view({"get": "retrieve"}),
        name="entry-featured",
    ),
    url(
        r"^authors/(?P<pk>[^/.]+)/(?P<related_field>\w+)/$",
        AuthorViewSet.as_view({"get": "retrieve_related"}),
        name="author-related",
    ),
    url(
        r"^entries/(?P<pk>[^/.]+)/relationships/(?P<related_field>\w+)$",
        EntryRelationshipView.as_view(),
        name="entry-relationships",
    ),
    url(
        r"^blogs/(?P<pk>[^/.]+)/relationships/(?P<related_field>\w+)$",
        BlogRelationshipView.as_view(),
        name="blog-relationships",
    ),
    url(
        r"^comments/(?P<pk>[^/.]+)/relationships/(?P<related_field>\w+)$",
        CommentRelationshipView.as_view(),
        name="comment-relationships",
    ),
    url(
        r"^authors/(?P<pk>[^/.]+)/relationships/(?P<related_field>\w+)$",
        AuthorRelationshipView.as_view(),
        name="author-relationships",
    ),
    path(
        "openapi",
        get_schema_view(
            title="Example API",
            description="API for all things …",
            version="1.0.0",
            generator_class=SchemaGenerator,
        ),
        name="openapi-schema",
    ),
    path(
        "swagger-ui/",
        TemplateView.as_view(
            template_name="swagger-ui.html",
            extra_context={"schema_url": "openapi-schema"},
        ),
        name="swagger-ui",
    ),
]

if settings.DEBUG:
    import debug_toolbar

    urlpatterns = [
        url(r"^__debug__/", include(debug_toolbar.urls)),
    ] + urlpatterns



#!/usr/bin/env python3

import os
import re
import sys

from setuptools import setup

needs_wheel = {"bdist_wheel"}.intersection(sys.argv)
wheel = ["wheel"] if needs_wheel else []


def read(*paths):
    """
    Build a file path from paths and return the contents.
    """
    with open(os.path.join(*paths), "r") as f:
        return f.read()


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, "__init__.py")).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [
        dirpath
        for dirpath, dirnames, filenames in os.walk(package)
        if os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [
        (dirpath.replace(package + os.sep, "", 1), filenames)
        for dirpath, dirnames, filenames in os.walk(package)
        if not os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename) for filename in filenames])
    return {package: filepaths}


if sys.argv[-1] == "publish":
    os.system("python setup.py sdist upload")
    os.system("python setup.py bdist_wheel upload")
    print("You probably want to also tag the version now:")
    print(
        "  git tag -a {0} -m 'version {0}'".format(
            get_version("rest_framework_json_api")
        )
    )
    print("  git push --tags")
    sys.exit()

setup(
    name="djangorestframework-jsonapi",
    version=get_version("rest_framework_json_api"),
    url="https://github.com/django-json-api/django-rest-framework-json-api",
    license="BSD",
    description="A Django REST framework API adapter for the JSON:API spec.",
    long_description=read("README.rst"),
    author="Jerel Unruh",
    author_email="",
    packages=get_packages("rest_framework_json_api"),
    package_data=get_package_data("rest_framework_json_api"),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=[
        "inflection>=0.3.0",
        "djangorestframework>=3.12,<3.13",
        "django>=2.2,<4.1",
    ],
    extras_require={
        "django-polymorphic": ["django-polymorphic>=2.0"],
        "django-filter": ["django-filter>=2.0"],
        "openapi": ["pyyaml>=5.3", "uritemplate>=3.0.1"],
    },
    setup_requires=wheel,
    python_requires=">=3.6",
    zip_safe=False,
)



import warnings
from collections import OrderedDict
from collections.abc import Mapping

import inflection
from django.core.exceptions import ObjectDoesNotExist
from django.db.models.query import QuerySet
from django.utils.module_loading import import_string as import_class_from_dotted_path
from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import ParseError

# star import defined so `rest_framework_json_api.serializers` can be
# a simple drop in for `rest_framework.serializers`
from rest_framework.serializers import *  # noqa: F401, F403
from rest_framework.serializers import (
    BaseSerializer,
    HyperlinkedModelSerializer,
    ModelSerializer,
    Serializer,
    SerializerMetaclass,
)
from rest_framework.settings import api_settings

from rest_framework_json_api.exceptions import Conflict
from rest_framework_json_api.relations import ResourceRelatedField
from rest_framework_json_api.utils import (
    get_included_resources,
    get_resource_type_from_instance,
    get_resource_type_from_model,
    get_resource_type_from_serializer,
)


class ResourceIdentifierObjectSerializer(BaseSerializer):
    default_error_messages = {
        "incorrect_model_type": _(
            "Incorrect model type. Expected {model_type}, received {received_type}."
        ),
        "does_not_exist": _('Invalid pk "{pk_value}" - object does not exist.'),
        "incorrect_type": _("Incorrect type. Expected pk value, received {data_type}."),
    }

    model_class = None

    def __init__(self, *args, **kwargs):
        self.model_class = kwargs.pop("model_class", self.model_class)
        # this has no fields but assumptions are made elsewhere that self.fields exists.
        self.fields = {}
        super().__init__(*args, **kwargs)

    def to_representation(self, instance):
        return {
            "type": get_resource_type_from_instance(instance),
            "id": str(instance.pk),
        }

    def to_internal_value(self, data):
        if data["type"] != get_resource_type_from_model(self.model_class):
            self.fail(
                "incorrect_model_type",
                model_type=self.model_class,
                received_type=data["type"],
            )
        pk = data["id"]
        try:
            return self.model_class.objects.get(pk=pk)
        except ObjectDoesNotExist:
            self.fail("does_not_exist", pk_value=pk)
        except (TypeError, ValueError):
            self.fail("incorrect_type", data_type=type(data["pk"]).__name__)


class SparseFieldsetsMixin:
    """
    A serializer mixin that adds support for sparse fieldsets through `fields` query parameter.

    Specification: https://jsonapi.org/format/#fetching-sparse-fieldsets
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        context = kwargs.get("context")
        request = context.get("request") if context else None

        if request:
            sparse_fieldset_query_param = "fields[{}]".format(
                get_resource_type_from_serializer(self)
            )
            try:
                param_name = next(
                    key
                    for key in request.query_params
                    if sparse_fieldset_query_param == key
                )
            except StopIteration:
                pass
            else:
                fieldset = request.query_params.get(param_name).split(",")
                # iterate over a *copy* of self.fields' underlying OrderedDict, because we may
                # modify the original during the iteration.
                # self.fields is a `rest_framework.utils.serializer_helpers.BindingDict`
                for field_name, field in self.fields.fields.copy().items():
                    if (
                        field_name == api_settings.URL_FIELD_NAME
                    ):  # leave self link there
                        continue
                    if field_name not in fieldset:
                        self.fields.pop(field_name)


class IncludedResourcesValidationMixin:
    """
    A serializer mixin that adds validation of `include` query parameter to
    support compound documents.

    Specification: https://jsonapi.org/format/#document-compound-documents)
    """

    def __init__(self, *args, **kwargs):
        context = kwargs.get("context")
        request = context.get("request") if context else None
        view = context.get("view") if context else None

        def validate_path(serializer_class, field_path, path):
            serializers = getattr(serializer_class, "included_serializers", None)
            if serializers is None:
                raise ParseError("This endpoint does not support the include parameter")
            this_field_name = inflection.underscore(field_path[0])
            this_included_serializer = serializers.get(this_field_name)
            if this_included_serializer is None:
                raise ParseError(
                    "This endpoint does not support the include parameter for path {}".format(
                        path
                    )
                )
            if len(field_path) > 1:
                new_included_field_path = field_path[1:]
                # We go down one level in the path
                validate_path(this_included_serializer, new_included_field_path, path)

        if request and view:
            included_resources = get_included_resources(request)
            for included_field_name in included_resources:
                included_field_path = included_field_name.split(".")
                if "related_field" in view.kwargs:
                    this_serializer_class = view.get_related_serializer_class()
                else:
                    this_serializer_class = view.get_serializer_class()
                # lets validate the current path
                validate_path(
                    this_serializer_class, included_field_path, included_field_name
                )

        super().__init__(*args, **kwargs)


class ReservedFieldNamesMixin:
    """Ensures that reserved field names are not used and an error raised instead."""

    _reserved_field_names = {"meta", "results"}

    def get_fields(self):
        fields = super().get_fields()

        found_reserved_field_names = self._reserved_field_names.intersection(
            fields.keys()
        )
        assert not found_reserved_field_names, (
            f"Serializer class {self.__class__.__module__}.{self.__class__.__qualname__} "
            f"uses following reserved field name(s) which is not allowed: "
            f"{', '.join(sorted(found_reserved_field_names))}"
        )

        if "type" in fields:
            # see https://jsonapi.org/format/#document-resource-object-fields
            warnings.warn(
                DeprecationWarning(
                    f"Field name 'type'  found in serializer class "
                    f"{self.__class__.__module__}.{self.__class__.__qualname__} "
                    f"which is not allowed according to the JSON:API spec and "
                    f"won't be supported anymore in the next major DJA release. "
                    f"Rename 'type' field to something else. "
                )
            )

        return fields


class LazySerializersDict(Mapping):
    """
    A dictionary of serializers which lazily import dotted class path and self.
    """

    def __init__(self, parent, serializers):
        self.parent = parent
        self.serializers = serializers

    def __getitem__(self, key):
        value = self.serializers[key]
        if not isinstance(value, type):
            if value == "self":
                value = self.parent
            else:
                value = import_class_from_dotted_path(value)
            self.serializers[key] = value

        return value

    def __iter__(self):
        return iter(self.serializers)

    def __len__(self):
        return len(self.serializers)

    def __repr__(self):
        return dict.__repr__(self.serializers)


class SerializerMetaclass(SerializerMetaclass):
    def __new__(cls, name, bases, attrs):
        serializer = super().__new__(cls, name, bases, attrs)

        if attrs.get("included_serializers", None):
            setattr(
                serializer,
                "included_serializers",
                LazySerializersDict(serializer, attrs["included_serializers"]),
            )

        if attrs.get("related_serializers", None):
            setattr(
                serializer,
                "related_serializers",
                LazySerializersDict(serializer, attrs["related_serializers"]),
            )

        return serializer


# If user imports serializer from here we can catch class definition and check
# nested serializers for depricated use.
class Serializer(
    IncludedResourcesValidationMixin,
    SparseFieldsetsMixin,
    ReservedFieldNamesMixin,
    Serializer,
    metaclass=SerializerMetaclass,
):
    """
    A `Serializer` is a model-less serializer class with additional
    support for JSON:API spec features.

    As in JSON:API specification a type is always required you need to
    make sure that you define `resource_name` in your `Meta` class
    when deriving from this class.

    Included Mixins:

    * A mixin class to enable sparse fieldsets is included
    * A mixin class to enable validation of included resources is included
    """

    pass


class HyperlinkedModelSerializer(
    IncludedResourcesValidationMixin,
    SparseFieldsetsMixin,
    ReservedFieldNamesMixin,
    HyperlinkedModelSerializer,
    metaclass=SerializerMetaclass,
):
    """
    A type of `ModelSerializer` that uses hyperlinked relationships instead
    of primary key relationships. Specifically:

    * A 'url' field is included instead of the 'id' field.
    * Relationships to other instances are hyperlinks, instead of primary keys.

    Included Mixins:

    * A mixin class to enable sparse fieldsets is included
    * A mixin class to enable validation of included resources is included
    """


class ModelSerializer(
    IncludedResourcesValidationMixin,
    SparseFieldsetsMixin,
    ReservedFieldNamesMixin,
    ModelSerializer,
    metaclass=SerializerMetaclass,
):
    """
    A `ModelSerializer` is just a regular `Serializer`, except that:

    * A set of default fields are automatically populated.
    * A set of default validators are automatically populated.
    * Default `.create()` and `.update()` implementations are provided.

    The process of automatically determining a set of serializer fields
    based on the model fields is reasonably complex, but you almost certainly
    don't need to dig into the implementation.

    If the `ModelSerializer` class *doesn't* generate the set of fields that
    you need you should either declare the extra/differing fields explicitly on
    the serializer class, or simply use a `Serializer` class.


    Included Mixins:

    * A mixin class to enable sparse fieldsets is included
    * A mixin class to enable validation of included resources is included
    """

    serializer_related_field = ResourceRelatedField

    def get_field_names(self, declared_fields, info):
        """
        We override the parent to omit explicity defined meta fields (such
        as SerializerMethodFields) from the list of declared fields
        """
        meta_fields = getattr(self.Meta, "meta_fields", [])

        declared = OrderedDict()
        for field_name in set(declared_fields.keys()):
            field = declared_fields[field_name]
            if field_name not in meta_fields:
                declared[field_name] = field
        fields = super().get_field_names(declared, info)
        return list(fields) + list(getattr(self.Meta, "meta_fields", list()))


class PolymorphicSerializerMetaclass(SerializerMetaclass):
    """
    This metaclass ensures that the `polymorphic_serializers` is correctly defined on a
    `PolymorphicSerializer` class and make a cache of model/serializer/type mappings.
    """

    def __new__(cls, name, bases, attrs):
        new_class = super().__new__(cls, name, bases, attrs)

        # Ensure initialization is only performed for subclasses of PolymorphicModelSerializer
        # (excluding PolymorphicModelSerializer class itself).
        parents = [b for b in bases if isinstance(b, PolymorphicSerializerMetaclass)]
        if not parents:
            return new_class

        polymorphic_serializers = getattr(new_class, "polymorphic_serializers", None)
        if not polymorphic_serializers:
            raise NotImplementedError(
                "A PolymorphicModelSerializer must define a `polymorphic_serializers` attribute."
            )
        serializer_to_model = {
            serializer: serializer.Meta.model for serializer in polymorphic_serializers
        }
        model_to_serializer = {
            serializer.Meta.model: serializer for serializer in polymorphic_serializers
        }
        type_to_serializer = {
            get_resource_type_from_serializer(serializer): serializer
            for serializer in polymorphic_serializers
        }
        new_class._poly_serializer_model_map = serializer_to_model
        new_class._poly_model_serializer_map = model_to_serializer
        new_class._poly_type_serializer_map = type_to_serializer
        new_class._poly_force_type_resolution = True

        # Flag each linked polymorphic serializer to force type resolution based on instance
        for serializer in polymorphic_serializers:
            serializer._poly_force_type_resolution = True

        return new_class


class PolymorphicModelSerializer(
    ModelSerializer, metaclass=PolymorphicSerializerMetaclass
):
    """
    A serializer for polymorphic models.
    Useful for "lazy" parent models. Leaves should be represented with a regular serializer.
    """

    def get_fields(self):
        """
        Return an exhaustive list of the polymorphic serializer fields.
        """
        if self.instance not in (None, []):
            if not isinstance(self.instance, QuerySet):
                serializer_class = self.get_polymorphic_serializer_for_instance(
                    self.instance
                )
                return serializer_class(
                    self.instance, context=self.context
                ).get_fields()
            else:
                raise Exception(
                    "Cannot get fields from a polymorphic serializer given a queryset"
                )
        return super().get_fields()

    @classmethod
    def get_polymorphic_serializer_for_instance(cls, instance):
        """
        Return the polymorphic serializer associated with the given instance/model.
        Raise `NotImplementedError` if no serializer is found for the given model. This usually
        means that a serializer is missing in the class's `polymorphic_serializers` attribute.
        """
        try:
            return cls._poly_model_serializer_map[instance._meta.model]
        except KeyError:
            raise NotImplementedError(
                "No polymorphic serializer has been found for model {}".format(
                    instance._meta.model.__name__
                )
            )

    @classmethod
    def get_polymorphic_model_for_serializer(cls, serializer):
        """
        Return the polymorphic model associated with the given serializer.
        Raise `NotImplementedError` if no model is found for the given serializer. This usually
        means that a serializer is missing in the class's `polymorphic_serializers` attribute.
        """
        try:
            return cls._poly_serializer_model_map[serializer]
        except KeyError:
            raise NotImplementedError(
                "No polymorphic model has been found for serializer {}".format(
                    serializer.__name__
                )
            )

    @classmethod
    def get_polymorphic_serializer_for_type(cls, obj_type):
        """
        Return the polymorphic serializer associated with the given type.
        Raise `NotImplementedError` if no serializer is found for the given type. This usually
        means that a serializer is missing in the class's `polymorphic_serializers` attribute.
        """
        try:
            return cls._poly_type_serializer_map[obj_type]
        except KeyError:
            raise NotImplementedError(
                "No polymorphic serializer has been found for type {}".format(obj_type)
            )

    @classmethod
    def get_polymorphic_model_for_type(cls, obj_type):
        """
        Return the polymorphic model associated with the given type.
        Raise `NotImplementedError` if no model is found for the given type. This usually
        means that a serializer is missing in the class's `polymorphic_serializers` attribute.
        """
        return cls.get_polymorphic_model_for_serializer(
            cls.get_polymorphic_serializer_for_type(obj_type)
        )

    @classmethod
    def get_polymorphic_types(cls):
        """
        Return the list of accepted types.
        """
        return cls._poly_type_serializer_map.keys()

    def to_representation(self, instance):
        """
        Retrieve the appropriate polymorphic serializer and use this to handle representation.
        """
        serializer_class = self.get_polymorphic_serializer_for_instance(instance)
        return serializer_class(instance, context=self.context).to_representation(
            instance
        )

    def to_internal_value(self, data):
        """
        Ensure that the given type is one of the expected polymorphic types, then retrieve the
        appropriate polymorphic serializer and use this to handle internal value.
        """
        received_type = data.get("type")
        expected_types = self.get_polymorphic_types()
        if received_type not in expected_types:
            raise Conflict(
                "Incorrect relation type. Expected on of [{expected_types}], "
                "received {received_type}.".format(
                    expected_types=", ".join(expected_types),
                    received_type=received_type,
                )
            )
        serializer_class = self.get_polymorphic_serializer_for_type(received_type)
        self.__class__ = serializer_class
        return serializer_class(
            self.instance, data, context=self.context, partial=self.partial
        ).to_internal_value(data)



from rest_framework_json_api.views import ModelViewSet
from tests.models import BasicModel
from tests.serializers import BasicModelSerializer


class BasicModelViewSet(ModelViewSet):
    serializer_class = BasicModelSerializer

    class Meta:
        model = BasicModel



import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("example", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="TaggedItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                ("tag", models.SlugField()),
                ("object_id", models.PositiveIntegerField()),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.ContentType",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



import re

import pytest
from django.urls import reverse

pytestmark = pytest.mark.django_db


def test_browsable_api_with_included_serializers(single_entry, client):
    response = client.get(
        reverse("entry-detail", kwargs={"pk": single_entry.pk, "format": "api"})
    )
    content = str(response.content)
    assert response.status_code == 200
    assert re.search(r"JSON:API includes", content)
    assert re.search(
        r'<input type="checkbox" name="includes" [^>]* value="authors.bio"', content
    )


def test_browsable_api_on_related_url(author, client):
    url = reverse("author-related", kwargs={"pk": author.pk, "related_field": "bio"})
    response = client.get(url, data={"format": "api"})
    content = str(response.content)
    assert response.status_code == 200
    assert re.search(r"JSON:API includes", content)
    assert re.search(
        r'<input type="checkbox" name="includes" [^>]* value="metadata"', content
    )


def test_browsable_api_with_no_included_serializers(client):
    response = client.get(reverse("projecttype-list", kwargs={"format": "api"}))
    content = str(response.content)
    assert response.status_code == 200
    assert not re.search(r"JSON:API includes", content)



import pytest
from django.urls import reverse

pytestmark = pytest.mark.django_db


def test_included_data_on_list(multiple_entries, client):
    response = client.get(
        reverse("entry-list"), data={"include": "comments", "page[size]": 5}
    )
    included = response.json().get("included")

    assert len(response.json()["data"]) == len(
        multiple_entries
    ), "Incorrect entry count"
    assert [x.get("type") for x in included] == [
        "comments",
        "comments",
    ], "List included types are incorrect"

    comment_count = len(
        [resource for resource in included if resource["type"] == "comments"]
    )
    expected_comment_count = sum(entry.comments.count() for entry in multiple_entries)
    assert comment_count == expected_comment_count, "List comment count is incorrect"


def test_included_data_on_list_with_one_to_one_relations(multiple_entries, client):
    response = client.get(
        reverse("entry-list"), data={"include": "authors.bio.metadata", "page[size]": 5}
    )
    included = response.json().get("included")

    assert len(response.json()["data"]) == len(
        multiple_entries
    ), "Incorrect entry count"
    expected_include_types = [
        "authorBioMetadata",
        "authorBioMetadata",
        "authorBios",
        "authorBios",
        "authors",
        "authors",
    ]
    include_types = [x.get("type") for x in included]
    assert include_types == expected_include_types, "List included types are incorrect"


def test_default_included_data_on_detail(single_entry, client):
    return test_included_data_on_detail(
        single_entry=single_entry, client=client, query=""
    )


def test_included_data_on_detail(single_entry, client, query="?include=comments"):
    response = client.get(
        reverse("entry-detail", kwargs={"pk": single_entry.pk}) + query
    )
    included = response.json().get("included")

    assert [x.get("type") for x in included] == [
        "comments"
    ], "Detail included types are incorrect"

    comment_count = len(
        [resource for resource in included if resource["type"] == "comments"]
    )
    expected_comment_count = single_entry.comments.count()
    assert comment_count == expected_comment_count, "Detail comment count is incorrect"


def test_dynamic_related_data_is_included(single_entry, entry_factory, client):
    entry_factory()
    response = client.get(
        reverse("entry-detail", kwargs={"pk": single_entry.pk}) + "?include=featured"
    )
    included = response.json().get("included")

    assert [x.get("type") for x in included] == [
        "entries"
    ], "Dynamic included types are incorrect"
    assert (
        len(included) == 1
    ), "The dynamically included blog entries are of an incorrect count"


def test_dynamic_many_related_data_is_included(single_entry, entry_factory, client):
    entry_factory()
    response = client.get(
        reverse("entry-detail", kwargs={"pk": single_entry.pk}) + "?include=suggested"
    )
    included = response.json().get("included")

    assert included
    assert [x.get("type") for x in included] == [
        "entries"
    ], "Dynamic included types are incorrect"


def test_missing_field_not_included(author_bio_factory, author_factory, client):
    # First author does not have a bio
    author = author_factory(bio=None)
    response = client.get(reverse("author-detail", args=[author.pk]) + "?include=bio")
    assert "included" not in response.json()
    # Second author does
    author = author_factory()
    response = client.get(reverse("author-detail", args=[author.pk]) + "?include=bio")
    data = response.json()
    assert "included" in data
    assert len(data["included"]) == 1
    assert data["included"][0]["attributes"]["body"] == author.bio.body


def test_deep_included_data_on_list(multiple_entries, client):
    response = client.get(
        reverse("entry-list") + "?include=comments,comments.author,"
        "comments.author.bio,comments.writer&page[size]=5"
    )
    included = response.json().get("included")

    assert len(response.json()["data"]) == len(
        multiple_entries
    ), "Incorrect entry count"
    assert [x.get("type") for x in included] == [
        "authorBios",
        "authorBios",
        "authors",
        "authors",
        "comments",
        "comments",
        "writers",
        "writers",
    ], "List included types are incorrect"

    comment_count = len(
        [resource for resource in included if resource["type"] == "comments"]
    )
    expected_comment_count = sum(entry.comments.count() for entry in multiple_entries)
    assert comment_count == expected_comment_count, "List comment count is incorrect"

    author_count = len(
        [resource for resource in included if resource["type"] == "authors"]
    )
    expected_author_count = sum(
        entry.comments.filter(author__isnull=False).count()
        for entry in multiple_entries
    )
    assert author_count == expected_author_count, "List author count is incorrect"

    author_bio_count = len(
        [resource for resource in included if resource["type"] == "authorBios"]
    )
    expected_author_bio_count = sum(
        entry.comments.filter(author__bio__isnull=False).count()
        for entry in multiple_entries
    )
    assert (
        author_bio_count == expected_author_bio_count
    ), "List author bio count is incorrect"

    writer_count = len(
        [resource for resource in included if resource["type"] == "writers"]
    )
    expected_writer_count = sum(
        entry.comments.filter(author__isnull=False).count()
        for entry in multiple_entries
    )
    assert writer_count == expected_writer_count, "List writer count is incorrect"

    # Also include entry authors
    response = client.get(
        reverse("entry-list") + "?include=authors,comments,comments.author,"
        "comments.author.bio&page[size]=5"
    )
    included = response.json().get("included")

    assert len(response.json()["data"]) == len(
        multiple_entries
    ), "Incorrect entry count"
    assert [x.get("type") for x in included] == [
        "authorBios",
        "authorBios",
        "authors",
        "authors",
        "authors",
        "authors",
        "comments",
        "comments",
    ], "List included types are incorrect"

    author_count = len(
        [resource for resource in included if resource["type"] == "authors"]
    )
    expected_author_count = sum(
        [entry.authors.count() for entry in multiple_entries]
        + [
            entry.comments.filter(author__isnull=False).count()
            for entry in multiple_entries
        ]
    )
    assert author_count == expected_author_count, "List author count is incorrect"


def test_deep_included_data_on_detail(single_entry, client):
    # Same test as in list but also ensures that intermediate resources (here comments' authors)
    # are returned along with the leaf nodes
    response = client.get(
        reverse("entry-detail", kwargs={"pk": single_entry.pk})
        + "?include=comments,comments.author.bio"
    )
    included = response.json().get("included")

    assert [x.get("type") for x in included] == [
        "authorBios",
        "authors",
        "comments",
    ], "Detail included types are incorrect"

    comment_count = len(
        [resource for resource in included if resource["type"] == "comments"]
    )
    expected_comment_count = single_entry.comments.count()
    assert comment_count == expected_comment_count, "Detail comment count is incorrect"

    author_bio_count = len(
        [resource for resource in included if resource["type"] == "authorBios"]
    )
    expected_author_bio_count = single_entry.comments.filter(
        author__bio__isnull=False
    ).count()
    assert (
        author_bio_count == expected_author_bio_count
    ), "Detail author bio count is incorrect"


def test_data_resource_not_included_again(single_comment, client):
    # This test makes sure that the resource which is in the data field is excluded
    # from the included field.
    response = client.get(
        reverse("comment-detail", kwargs={"pk": single_comment.pk})
        + "?include=entry.comments"
    )

    included = response.json().get("included")

    included_comments = [
        resource for resource in included if resource["type"] == "comments"
    ]
    assert single_comment.pk not in [
        int(x.get("id")) for x in included_comments
    ], "Resource of the data field duplicated in included"

    comment_count = len(included_comments)
    expected_comment_count = single_comment.entry.comments.count()
    # The comment in the data attribute must not be included again.
    expected_comment_count -= 1
    assert comment_count == expected_comment_count, "Comment count incorrect"


def test_meta_object_added_to_included_resources(single_entry, client):
    response = client.get(
        reverse("entry-detail", kwargs={"pk": single_entry.pk}) + "?include=comments"
    )
    assert response.json()["included"][0].get("meta")

    response = client.get(
        reverse("entry-detail", kwargs={"pk": single_entry.pk})
        + "?include=comments.author"
    )
    assert response.json()["included"][0].get("meta")
    assert response.json()["included"][1].get("meta")



import pytest
from django.contrib.auth import get_user_model
from django.test import override_settings
from django.urls import reverse
from django.utils import encoding

from example.tests import TestBase


class ModelViewSetTests(TestBase):
    """
    Test usage with ModelViewSets, also tests pluralization, camelization,
    and underscore.

    [<RegexURLPattern user-list ^identities/$>,
    <RegexURLPattern user-detail ^identities/(?P<pk>[^/]+)/$>]
    """

    list_url = reverse("user-list")

    def setUp(self):
        super().setUp()
        self.detail_url = reverse("user-detail", kwargs={"pk": self.miles.pk})

    def test_key_in_list_result(self):
        """
        Ensure the result has a 'user' key since that is the name of the model
        """
        with override_settings(JSON_API_FORMAT_FIELD_NAMES="dasherize"):
            response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)

        user = get_user_model().objects.all()[0]
        expected = {
            "data": [
                {
                    "type": "users",
                    "id": encoding.force_str(user.pk),
                    "attributes": {
                        "first-name": user.first_name,
                        "last-name": user.last_name,
                        "email": user.email,
                    },
                }
            ],
            "links": {
                "first": "http://testserver/identities?page%5Bnumber%5D=1",
                "last": "http://testserver/identities?page%5Bnumber%5D=2",
                "next": "http://testserver/identities?page%5Bnumber%5D=2",
                "prev": None,
            },
            "meta": {"pagination": {"page": 1, "pages": 2, "count": 2}},
        }

        assert expected == response.json()

    def test_page_two_in_list_result(self):
        """
        Ensure that the second page is reachable and is the correct data.
        """
        with override_settings(JSON_API_FORMAT_FIELD_NAMES="dasherize"):
            response = self.client.get(self.list_url, {"page[number]": 2})
        self.assertEqual(response.status_code, 200)

        user = get_user_model().objects.all()[1]
        expected = {
            "data": [
                {
                    "type": "users",
                    "id": encoding.force_str(user.pk),
                    "attributes": {
                        "first-name": user.first_name,
                        "last-name": user.last_name,
                        "email": user.email,
                    },
                }
            ],
            "links": {
                "first": "http://testserver/identities?page%5Bnumber%5D=1",
                "last": "http://testserver/identities?page%5Bnumber%5D=2",
                "next": None,
                "prev": "http://testserver/identities?page%5Bnumber%5D=1",
            },
            "meta": {"pagination": {"page": 2, "pages": 2, "count": 2}},
        }

        assert expected == response.json()

    def test_page_range_in_list_result(self):
        """
        Ensure that the range of a page can be changed from the client,
        tests pluralization as two objects means it converts ``user`` to
        ``users``.
        """
        with override_settings(JSON_API_FORMAT_FIELD_NAMES="dasherize"):
            response = self.client.get(self.list_url, {"page[size]": 2})
        self.assertEqual(response.status_code, 200)

        users = get_user_model().objects.all()
        expected = {
            "data": [
                {
                    "type": "users",
                    "id": encoding.force_str(users[0].pk),
                    "attributes": {
                        "first-name": users[0].first_name,
                        "last-name": users[0].last_name,
                        "email": users[0].email,
                    },
                },
                {
                    "type": "users",
                    "id": encoding.force_str(users[1].pk),
                    "attributes": {
                        "first-name": users[1].first_name,
                        "last-name": users[1].last_name,
                        "email": users[1].email,
                    },
                },
            ],
            "links": {
                "first": "http://testserver/identities?page%5Bnumber%5D=1&page%5Bsize%5D=2",
                "last": "http://testserver/identities?page%5Bnumber%5D=1&page%5Bsize%5D=2",
                "next": None,
                "prev": None,
            },
            "meta": {"pagination": {"page": 1, "pages": 1, "count": 2}},
        }

        assert expected == response.json()

    def test_key_in_detail_result(self):
        """
        Ensure the result has a 'user' key.
        """
        with override_settings(JSON_API_FORMAT_FIELD_NAMES="dasherize"):
            response = self.client.get(self.detail_url)
        self.assertEqual(response.status_code, 200)

        expected = {
            "data": {
                "type": "users",
                "id": encoding.force_str(self.miles.pk),
                "attributes": {
                    "first-name": self.miles.first_name,
                    "last-name": self.miles.last_name,
                    "email": self.miles.email,
                },
            }
        }

        assert expected == response.json()

    def test_patch_requires_id(self):
        """
        Verify that 'id' is required to be passed in an update request.
        """
        data = {
            "data": {"type": "users", "attributes": {"first-name": "DifferentName"}}
        }

        response = self.client.patch(self.detail_url, data=data)

        self.assertEqual(response.status_code, 400)

    def test_patch_requires_correct_id(self):
        """
        Verify that 'id' is the same then in url
        """
        data = {
            "data": {
                "type": "users",
                "id": self.miles.pk + 1,
                "attributes": {"first-name": "DifferentName"},
            }
        }

        response = self.client.patch(self.detail_url, data=data)

        self.assertEqual(response.status_code, 409)

    def test_key_in_post(self):
        """
        Ensure a key is in the post.
        """
        self.client.login(username="miles", password="pw")
        data = {
            "data": {
                "type": "users",
                "id": encoding.force_str(self.miles.pk),
                "attributes": {
                    "first-name": self.miles.first_name,
                    "last-name": self.miles.last_name,
                    "email": "miles@trumpet.org",
                },
            }
        }

        with override_settings(JSON_API_FORMAT_FIELD_NAMES="dasherize"):
            response = self.client.put(self.detail_url, data=data)

        assert data == response.json()

        # is it updated?
        self.assertEqual(
            get_user_model().objects.get(pk=self.miles.pk).email, "miles@trumpet.org"
        )

    def test_404_error_pointer(self):
        self.client.login(username="miles", password="pw")
        not_found_url = reverse("user-detail", kwargs={"pk": 12345})
        errors = {
            "errors": [{"detail": "Not found.", "status": "404", "code": "not_found"}]
        }

        response = self.client.get(not_found_url)
        assert 404 == response.status_code
        assert errors == response.json()


@pytest.mark.django_db
def test_patch_allow_field_type(author, author_type_factory, client):
    """
    Verify that type field may be updated.
    """
    # TODO remove in next major version 5.0.0 see serializers.ReservedFieldNamesMixin
    with pytest.deprecated_call():
        author_type = author_type_factory()
        url = reverse("author-detail", args=[author.id])

        data = {
            "data": {
                "id": author.id,
                "type": "authors",
                "relationships": {
                    "data": {"id": author_type.id, "type": "author-type"}
                },
            }
        }

        response = client.patch(url, data=data)

        assert response.status_code == 200



import pytest
from django.test import override_settings
from django.urls import path, reverse
from rest_framework import generics

from rest_framework_json_api import serializers

from example.models import Author, Blog


# serializers
class CommentAttachmentSerializer(serializers.Serializer):
    data = serializers.CharField(allow_null=False, required=True)

    def validate_data(self, value):
        if value and len(value) < 10:
            raise serializers.ValidationError("Too short data")


class CommentSerializer(serializers.Serializer):
    attachments = CommentAttachmentSerializer(many=True, required=False)
    attachment = CommentAttachmentSerializer(required=False)
    one_more_attachment = CommentAttachmentSerializer(required=False)
    body = serializers.CharField(allow_null=False, required=True)


class EntrySerializer(serializers.Serializer):
    blog = serializers.IntegerField()
    comments = CommentSerializer(many=True, required=False)
    comment = CommentSerializer(required=False)
    headline = serializers.CharField(allow_null=True, required=True)
    body_text = serializers.CharField()
    author = serializers.ResourceRelatedField(
        queryset=Author.objects.all(), required=False
    )

    def validate(self, attrs):
        body_text = attrs["body_text"]
        if len(body_text) < 5:
            raise serializers.ValidationError(
                {"body_text": {"title": "Too Short title", "detail": "Too short"}}
            )


# view
class DummyTestView(generics.CreateAPIView):
    serializer_class = EntrySerializer
    resource_name = "entries"

    def get_serializer_context(self):
        return {}


urlpatterns = [
    path("entries-nested", DummyTestView.as_view(), name="entries-nested-list")
]


@pytest.fixture(scope="function")
def some_blog(db):
    return Blog.objects.create(name="Some Blog", tagline="It's a blog")


def perform_error_test(client, data):
    with override_settings(ROOT_URLCONF=__name__):
        url = reverse("entries-nested-list")
        response = client.post(url, data=data)

    return response.json()


def test_first_level_attribute_error(client, some_blog, snapshot):
    data = {
        "data": {
            "type": "entries",
            "attributes": {
                "blog": some_blog.pk,
                "bodyText": "body_text",
            },
        }
    }
    assert snapshot == perform_error_test(client, data)


def test_first_level_custom_attribute_error(client, some_blog, snapshot):
    data = {
        "data": {
            "type": "entries",
            "attributes": {
                "blog": some_blog.pk,
                "body-text": "body",
                "headline": "headline",
            },
        }
    }
    with override_settings(JSON_API_FORMAT_FIELD_NAMES="dasherize"):
        assert snapshot == perform_error_test(client, data)


def test_second_level_array_error(client, some_blog, snapshot):
    data = {
        "data": {
            "type": "entries",
            "attributes": {
                "blog": some_blog.pk,
                "bodyText": "body_text",
                "headline": "headline",
                "comments": [{}],
            },
        }
    }

    assert snapshot == perform_error_test(client, data)


def test_second_level_dict_error(client, some_blog, snapshot):
    data = {
        "data": {
            "type": "entries",
            "attributes": {
                "blog": some_blog.pk,
                "bodyText": "body_text",
                "headline": "headline",
                "comment": {},
            },
        }
    }

    assert snapshot == perform_error_test(client, data)


def test_third_level_array_error(client, some_blog, snapshot):
    data = {
        "data": {
            "type": "entries",
            "attributes": {
                "blog": some_blog.pk,
                "bodyText": "body_text",
                "headline": "headline",
                "comments": [{"body": "test comment", "attachments": [{}]}],
            },
        }
    }

    assert snapshot == perform_error_test(client, data)


def test_third_level_custom_array_error(client, some_blog, snapshot):
    data = {
        "data": {
            "type": "entries",
            "attributes": {
                "blog": some_blog.pk,
                "bodyText": "body_text",
                "headline": "headline",
                "comments": [
                    {"body": "test comment", "attachments": [{"data": "text"}]}
                ],
            },
        }
    }

    assert snapshot == perform_error_test(client, data)


def test_third_level_dict_error(client, some_blog, snapshot):
    data = {
        "data": {
            "type": "entries",
            "attributes": {
                "blog": some_blog.pk,
                "bodyText": "body_text",
                "headline": "headline",
                "comments": [{"body": "test comment", "attachment": {}}],
            },
        }
    }

    assert snapshot == perform_error_test(client, data)


def test_many_third_level_dict_errors(client, some_blog, snapshot):
    data = {
        "data": {
            "type": "entries",
            "attributes": {
                "blog": some_blog.pk,
                "bodyText": "body_text",
                "headline": "headline",
                "comments": [{"attachment": {}}],
            },
        }
    }

    assert snapshot == perform_error_test(client, data)


def test_relationship_errors_has_correct_pointers(client, some_blog, snapshot):
    data = {
        "data": {
            "type": "entries",
            "attributes": {
                "blog": some_blog.pk,
                "bodyText": "body_text",
                "headline": "headline",
            },
            "relationships": {
                "author": {"data": {"id": "INVALID_ID", "type": "authors"}}
            },
        }
    }

    assert snapshot == perform_error_test(client, data)



from django.contrib.auth import get_user_model
from django.urls import reverse
from django.utils import encoding
from rest_framework import status

from example.tests import TestBase


class FormatKeysSetTests(TestBase):
    """
    Test that camelization and underscoring of key names works if they are activated.
    """

    list_url = reverse("user-list")

    def setUp(self):
        super().setUp()
        self.detail_url = reverse("user-detail", kwargs={"pk": self.miles.pk})

    def test_camelization(self):
        """
        Test that camelization works.
        """
        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)

        user = get_user_model().objects.all()[0]
        expected = {
            "data": [
                {
                    "type": "users",
                    "id": encoding.force_str(user.pk),
                    "attributes": {
                        "firstName": user.first_name,
                        "lastName": user.last_name,
                        "email": user.email,
                    },
                }
            ],
            "links": {
                "first": "http://testserver/identities?page%5Bnumber%5D=1",
                "last": "http://testserver/identities?page%5Bnumber%5D=2",
                "next": "http://testserver/identities?page%5Bnumber%5D=2",
                "prev": None,
            },
            "meta": {"pagination": {"page": 1, "pages": 2, "count": 2}},
        }

        assert expected == response.json()


def test_options_format_field_names(db, client):
    response = client.options(reverse("author-list"))
    assert response.status_code == status.HTTP_200_OK
    data = response.json()["data"]
    expected_keys = {
        "name",
        "email",
        "bio",
        "entries",
        "firstEntry",
        "type",
        "comments",
        "secrets",
        "defaults",
        "initials",
    }
    assert expected_keys == data["actions"]["POST"].keys()



from django.db import models


class DJAModel(models.Model):
    """
    Base for test models that sets app_label, so they play nicely.
    """

    class Meta:
        app_label = "tests"
        abstract = True


class BasicModel(DJAModel):
    text = models.CharField(max_length=100)


# Models for relations tests
# ManyToMany
class ManyToManyTarget(DJAModel):
    name = models.CharField(max_length=100)


class ManyToManySource(DJAModel):
    name = models.CharField(max_length=100)
    targets = models.ManyToManyField(ManyToManyTarget, related_name="sources")


# ForeignKey
class ForeignKeyTarget(DJAModel):
    name = models.CharField(max_length=100)


class ForeignKeySource(DJAModel):
    name = models.CharField(max_length=100)
    target = models.ForeignKey(
        ForeignKeyTarget, related_name="sources", on_delete=models.CASCADE
    )
